#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 13/11/2023, 14:18. Copyright (c) The Contributors

import os.path
import warnings
from typing import Union, Tuple, List

from astropy.units import Quantity

from .misc import cifbuild
from ..samples.base import BaseSample
from ..sources import BaseSource, GalaxyCluster
from ..sources.base import NullSource
from ..utils import RAD_LABELS, NUM_CORES
from ..exceptions import NotAssociatedError

def region_setup(sources: Union[BaseSource, BaseSample], outer_radius: Union[str, Quantity],
                 inner_radius: Union[str, Quantity], disable_progress: bool, obs_id: str, num_cores: int = NUM_CORES) \
        -> Tuple[Union[BaseSource, BaseSample], List[Quantity], List[Quantity]]:
    """
    The preparation and value checking stage for SAS spectrum generation.

    :param BaseSource/BaseSample sources: A single source object, or a sample of sources.
    :param str/Quantity outer_radius: The name or value of the outer radius to use for the generation of
        the spectrum (for instance 'r200' would be acceptable for a GalaxyCluster, or Quantity(1000, 'kpc')).
    :param str/Quantity inner_radius: The name or value of the inner radius to use for the generation of
        the spectrum (for instance 'r500' would be acceptable for a GalaxyCluster, or Quantity(300, 'kpc')). By
        default this is zero arcseconds, resulting in a circular spectrum.
    :param bool disable_progress: Setting this to true will turn off the SAS generation progress bar.
    :param str obs_id: Only used if the 'region' radius name is passed, the ObsID to retrieve the region for.
    :param int num_cores: The number of cores to be used, will be passed to cifbuild.
    :return: The source objects, a list of inner radius quantities, and a list of outer radius quantities.
    :rtype: Tuple[Union[BaseSource, BaseSample], List[Quantity], List[Quantity]]
    """
    # NullSources are not allowed to have spectra, as they can have any observations associated and thus won't
    #  necessarily overlap
    if isinstance(sources, NullSource):
        raise TypeError("You cannot create spectra of a NullSource")

    if isinstance(sources, BaseSource):
        sources = [sources]

    # Checking that the user hasn't passed BaseSources
    if not all([type(src) != BaseSource for src in sources]):
        raise TypeError("You cannot generate spectra from a BaseSource object, really you shouldn't be using "
                        "them at all, they are mostly useful as a superclass.")

    # Issuing a warning to the user that one or one sources have not been detected
    if not all([src.detected for src in sources]):
        warnings.warn("Not all of these sources have been detected, the spectra generated may not be helpful.")

    # Checking that inner radii that have been passed into the spectrum generation aren't nonsense
    if isinstance(inner_radius, str) and inner_radius not in RAD_LABELS:
        raise ValueError("You have passed a radius name rather than a value for 'inner_radius', but it is "
                         "not a valid name, please use one of the following:\n {}".format(", ".join(RAD_LABELS)))

    elif isinstance(inner_radius, str) and inner_radius in ["r2500", "r500", "r200"] and \
            not all([type(src) == GalaxyCluster for src in sources]):
        raise TypeError("The {} radius is only valid for GalaxyCluster objects".format(inner_radius))

    # One radius can be passed for a whole sample, but this checks to make sure that if there are multiple sources,
    #  and multiple radii have been passed, there are the same number of sources and radii
    elif isinstance(inner_radius, Quantity) and len(sources) != 1 and not inner_radius.isscalar \
            and len(sources) != len(inner_radius):
        raise ValueError("Your sample has {s} sources, but your inner_radius variable only has {i} entries. Please "
                         "pass only one inner_radius or the same number as there are "
                         "sources".format(s=len(sources), i=len(inner_radius)))

    # Checking that outer_radius radii that have been passed into the spectrum generation aren't nonsense
    if isinstance(outer_radius, str) and outer_radius not in RAD_LABELS:
        raise ValueError("You have passed a radius name rather than a value for 'outer_radius', but it is "
                         "not a valid name, please use one of the following:\n {}".format(", ".join(RAD_LABELS)))
    elif isinstance(outer_radius, str) and outer_radius in ["r2500", "r500", "r200"] and \
            not all([type(src) == GalaxyCluster for src in sources]):
        raise TypeError("The {} radius is only valid for GalaxyCluster objects".format(outer_radius))
    elif isinstance(outer_radius, Quantity) and len(sources) != 1 and not outer_radius.isscalar \
            and len(sources) != len(outer_radius):
        raise ValueError("Your sample has {s} sources, but your outer_radius variable only has {o} entries. Please "
                         "pass only one outer_radius or the same number as there are "
                         "sources".format(s=len(sources), o=len(outer_radius)))

    # A crude way to store the radii but I'm tired and this will work fine
    final_inner = []
    final_outer = []
    # I need to convert the radii to the same units and compare them, and to make sure they
    #  are actually in distance units. The distance unit checking is done by convert_radius
    for s_ind, src in enumerate(sources):
        # Converts the inner and outer radius for this source into the same unit
        if isinstance(outer_radius, str) and outer_radius != 'region':
            cur_out_rad = src.get_radius(outer_radius, 'deg')
        elif isinstance(outer_radius, str) and outer_radius == 'region':
            reg = src.source_back_regions('region', obs_id)[0]
            cur_out_rad = Quantity([reg.width.to('deg').value/2, reg.height.to('deg').value/2], 'deg')
        elif outer_radius.isscalar:
            cur_out_rad = src.convert_radius(outer_radius, 'deg')
        else:
            cur_out_rad = src.convert_radius(outer_radius[s_ind], 'deg')

        # We need to check that the outer radius isn't region, because for region objects we ignore whatever
        #  inner radius has been passed and just set it 0
        if outer_radius == 'region':
            cur_inn_rad = Quantity([0, 0], 'deg')
        elif isinstance(inner_radius, str):
            cur_inn_rad = src.get_radius(inner_radius, 'deg')
        elif inner_radius.isscalar:
            cur_inn_rad = src.convert_radius(inner_radius, 'deg')
        else:
            cur_inn_rad = src.convert_radius(inner_radius[s_ind], 'deg')

        # Then we can check to make sure that the outer radius is larger than the inner radius
        if outer_radius != 'region' and cur_inn_rad >= cur_out_rad:
            raise ValueError("The inner_radius of {s} is greater than or equal to the outer_radius".format(s=src.name))
        else:
            final_inner.append(cur_inn_rad)
            final_outer.append(cur_out_rad)

    # Have to make sure that all observations have an up to date cif file.
    cifbuild(sources, disable_progress=disable_progress, num_cores=num_cores)

    return sources, final_inner, final_outer


def check_pattern(pattern: Union[str, int]) -> Tuple[str, str]:
    """
    A very simple (and not exhaustive) checker for XMM SAS pattern expressions.

    :param str/int pattern: The pattern selection expression to be checked.
    :return: A string pattern selection expression, and a pattern representation that should be safe for naming
        SAS files with.
    :rtype: Tuple[str, str]
    """

    if isinstance(pattern, int):
        pattern = '==' + str(pattern)
    elif not isinstance(pattern, str):
        raise TypeError("Pattern arguments must be either an integer (we then assume only events with that pattern "
                        "should be selected) or a SAS selection command (e.g. 'in [1:4]' or '<= 4').")

    # First off I remove whitespace from the beginning and end of the term
    pattern = pattern.strip()
    # pattern = pattern.replace(' ', '')
    # Sometimes we will pass in patterns that have been converted to the 'XGA format', if you want to call it
    #  that. This namely happens when we're reading light curves that have been saved to disk back in. As such we
    #  replace the XGA-isms with their original string meanings
    pattern = pattern.replace('lteq', '<=').replace('gteq', '>=').replace('eq', '==') \
        .replace('lteq', '<=').replace('lt', '<').replace('gt', '>')

    # Then we check for understandable selection commands; inequalities, equals, and 'in'
    if pattern[:2] not in ['in', '<=', '>=', '=='] and pattern[:1] not in ['<', '>']:
        raise ValueError("First part of a pattern statement must be either 'in', '<=', '>=', '==', '<', or '>'.")

    if pattern[:2] == 'in' and '[' not in pattern and '(' not in pattern:
        raise ValueError("If a pattern statement uses 'in', either a '[' (for inclusive lower limit) or '(' (for "
                         "exclusive lower limit) must be in the statement.")

    if pattern[:2] == 'in' and ']' not in pattern and ')' not in pattern:
        raise ValueError("If a pattern statement uses 'in', either a ']' (for inclusive upper limit) or ')' (for "
                         "exclusive upper limit) must be in the statement.")

    if pattern[:2] == 'in' and ':' not in pattern:
        raise ValueError("If a pattern statement uses 'in', either a ':' must be present in the statement to separate "
                         "lower and upper limits.")

    # SAS doesn't like having file names with special characters, so I am trying to come up with safe replacements
    #  that still convey what the pattern selection was
    # .replace('in', '')
    patt_file_name = pattern.replace(' ', '').replace('<=', 'lteq').replace('>=', 'gteq').replace('==', 'eq')\
        .replace('<=', 'lteq').replace('<', 'lt').replace('>', 'gt')

    return pattern, patt_file_name


def _gen_detmap_cmd(source: BaseSource, obs_id: str, inst: str, bin_size: int = 200) -> Tuple[str, str, str]:
    """
    An internal method for generating SAS commands required to create detector maps for the weighting of ARFs.

    :param BaseSource source: The source for which the parent method is generating ARFs for, and that needs
        a detector map.
    :param str obs_id: The ObsID of the data we are generating ARFs for.
    :param str inst: The instrument of the data we are generating ARFs for. NOTE - ideally this instrument WILL NOT
        be used for the detector map, as it is beneficial to source a detector map from a different instrument to
        the one you are generating ARFs for.
    :param int bin_size: The x and y binning that should be applied to the image. Larger numbers will cause ARF
        generation to be faster, but arguably the results will be less accurate.
    :return: The command to generate the requested detector map (will be blank if the detector map already
        exists), the path where the detmap will be after the command is run (i.e. the ObsID directory if it was
        already generated, or the temporary directory if it has just been generated), and the final output path
        of the detector.
    :rtype: Tuple[str, str, str]
    """
    # This is the command that will be filled out to generate the detmap of our dreams!
    detmap_cmd = "evselect table={e} imageset={d} xcolumn=DETX ycolumn=DETY imagebinning=binSize ximagebinsize={bs} " \
                 "yimagebinsize={bs} {ex}"

    # Some settings depend on the instrument, we use different patterns for different instruments
    if "pn" in inst:
        # Also the upper channel limit is different for EPN and EMOS detectors
        spec_lim = 20479
        # This is an expression without region information to be used for making the detmaps
        #  required for ARF generation, we start off assuming we'll use a MOS observation as the detmap
        d_expr = "expression='#XMMEA_EM && (PATTERN <= 12) && (FLAG .eq. 0)'"

        # The detmap for the arfgen call should ideally not be from the same instrument as the observation,
        #  so for PN we preferentially select MOS2 (as MOS1 was damaged). However if there isn't a MOS2
        #  events list from the same observation then we select MOS1, and failing that we use PN.
        try:
            detmap_evts = source.get_products("events", obs_id=obs_id, inst='mos2')[0]
        except NotAssociatedError:
            try:
                detmap_evts = source.get_products("events", obs_id=obs_id, inst='mos1')[0]
            except NotAssociatedError:
                detmap_evts = source.get_products("events", obs_id=obs_id, inst='pn')[0]
                # If all is lost and there are no MOS event lists then we must revert to the PN expression
                d_expr = "expression='#XMMEA_EP && (PATTERN <= 4) && (FLAG .eq. 0)'"

    elif "mos" in inst:
        spec_lim = 11999
        # This is an expression without region information to be used for making the detmaps
        #  required for ARF generation, we start off assuming we'll use the PN observation as the detmap
        d_expr = "expression='#XMMEA_EP && (PATTERN <= 4) && (FLAG .eq. 0)'"

        # The detmap for the arfgen call should ideally not be from the same instrument as the observation,
        #  so for MOS observations we preferentially select PN. However if there isn't a PN events list
        #  from the same observation then for MOS2 we select MOS1, and for MOS1 we select MOS2 (as they
        #  are rotated wrt one another it is still semi-valid), and failing that MOS2 will get MOS2 and MOS1
        #  will get MOS1.
        if inst[-1] == 1:
            cur = '1'
            opp = '2'
        else:
            cur = '2'
            opp = '1'
        try:
            detmap_evts = source.get_products("events", obs_id=obs_id, inst='pn')[0]
        except NotAssociatedError:
            # If we must use a MOS detmap then we have to use the MOS expression
            d_expr = "expression='#XMMEA_EM && (PATTERN <= 12) && (FLAG .eq. 0)'"
            try:
                detmap_evts = source.get_products("events", obs_id=obs_id, inst='mos' + opp)[0]
            except NotAssociatedError:
                detmap_evts = source.get_products("events", obs_id=obs_id, inst='mos' + cur)[0]
    else:
        raise ValueError("You somehow have an illegal value for the instrument name...")

    det_map = "{o}_{i}_bin{bs}_detmap.fits".format(o=detmap_evts.obs_id, i=detmap_evts.instrument, bs=bin_size)
    det_map_path = os.path.join(OUTPUT, obs_id, det_map)

    # If the detmap that we've decided we need already exists, then we don't need to generate it again
    if os.path.exists(det_map_path):
        d_cmd_str = ""
        det_map_cmd_path = det_map_path
    else:
        # Does the same thing for the evselect command to make the detmap
        d_cmd_str = detmap_cmd.format(e=detmap_evts.path, d=det_map, ex=d_expr, bs=bin_size) + "; "
        det_map_cmd_path = det_map

    return d_cmd_str, det_map_cmd_path, det_map_path



#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 20/02/2023, 14:04. Copyright (c) The Contributors

from .base import BaseProduct, BaseAggregateProduct, BaseProfile1D, BaseAggregateProfile1D
from .misc import EventList
from .phot import Image, ExpMap, RateMap, PSF, PSFGrid
from .relation import ScalingRelation
from .spec import Spectrum, AnnularSpectra

# Defining a dictionary to map from string product names to their associated classes
PROD_MAP = {"image": Image, "expmap": ExpMap, "events": EventList, "spectrum": Spectrum, "psf": PSF,
            "psfgrid": PSFGrid}















#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 13/04/2023, 14:51. Copyright (c) The Contributors

from .clusters import *



#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 08/06/2023, 22:40. Copyright (c) The Contributors

from typing import Union, List

import numpy as np
from astropy.units import Quantity, Unit, UnitConversionError, kpc, deg
from scipy.special import gamma

from .base import BaseModel1D
from ..exceptions import XGAFitError
from ..utils import r500, r200, r2500


class BetaProfile1D(BaseModel1D):
    """
    An XGA model implementation of the beta profile, essentially a projected isothermal king profile, it can be
    used to describe a simple galaxy cluster radial surface brightness profile.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('ct/(s*arcmin**2)'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing the surface brightness beta profile model.
        """
        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('ct/(s*arcmin**2)'), Unit('ct/(s*kpc**2)'), Unit('ct/(s*pix**2)')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        poss_x_units = [kpc, deg, r200, r500, r2500]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        r_core_starts = [Quantity(100, 'kpc'), Quantity(0.2, 'deg'), Quantity(0.05, r200), Quantity(0.1, r500),
                         Quantity(0.5, r2500)]
        # TODO MAKE THE NEW START PARAMETERS MORE SENSIBLE
        norm_starts = [Quantity(1, 'ct/(s*arcmin**2)'), Quantity(1, 'ct/(s*kpc**2)'), Quantity(1, 'ct/(s*pix**2)')]

        start_pars = [Quantity(1, ''), r_core_starts[xu_ind], norm_starts[yu_ind]]
        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        # TODO ALSO MAKE THESE MORE SENSIBLE
        r_core_priors = [{'prior': Quantity([0, 2000], 'kpc'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], 'deg'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r200), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r500), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r2500), 'type': 'uniform'}]
        norm_priors = [{'prior': Quantity([0, 3], 'ct/(s*arcmin**2)'), 'type': 'uniform'},
                       {'prior': Quantity([0, 100], 'ct/(s*kpc**2)'), 'type': 'uniform'},
                       {'prior': Quantity([0, 100], 'ct/(s*pix**2)'), 'type': 'uniform'}]

        priors = [{'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind], norm_priors[yu_ind]]

        nice_pars = [r"$\beta$", r"R$_{\rm{core}}$", "S$_{0}$"]
        info_dict = {'author': 'placeholder', 'year': 'placeholder', 'reference': 'placeholder',
                     'general': 'Essentially a projected isothermal king profile, it can be\n'
                                'used to describe a simple galaxy cluster radial surface brightness profile.'}
        super().__init__(x_unit, y_unit, start_pars, priors, 'beta', 'Beta Profile', nice_pars, 'Surface Brightness',
                         info_dict)

    @staticmethod
    def model(x: Quantity, beta: Quantity, r_core: Quantity, norm: Quantity) -> Quantity:
        """
        The model function for the beta profile.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity beta: The beta slope parameter of the model.
        :param Quantity r_core: The core radius.
        :param Quantity norm: The normalisation of the model.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """
        return norm * ((1 + ((x / r_core)**2))**((-3 * beta) + 0.5))

    def derivative(self, x: Quantity, dx: Quantity = Quantity(0, ''), use_par_dist: bool = False) -> Quantity:
        """
        Calculates the gradient of the beta profile at a given point, overriding the numerical method implemented
        in the BaseModel1D class, as this simple model has an easily derivable first derivative.

        :param Quantity x: The point(s) at which the slope of the model should be measured.
        :param Quantity dx: This makes no difference here, as this is an analytical derivative. It has
            been left in so that the inputs for this method don't vary between models.
        :param bool use_par_dist: Should the parameter distributions be used to calculate a derivative
            distribution; this can only be used if a fit has been performed using the model instance.
            Default is False, in which case the current parameters will be used to calculate a single value.
        :return: The calculated slope of the model at the supplied x position(s).
        :rtype: Quantity
        """
        # Just makes sure that if there are multiple x values then the broadcasting will go to the correct shape of
        #  numpy array
        # This makes sure that the input radius, or radii, are being used properly. For a single x value, or a set of
        #  x values, we use [..., None] to ensure that (if the user has decided to use the parameter distribution) we
        #  get M distributions of the derivative. In the case where the user has passed a distribution of radii, really
        #  representative of a single radius but with uncertainty, then this does not trigger, and rather than an N
        #  by N (where N is the number of samples in each posterior) array, you get an M x N array
        if x.isscalar or (not x.isscalar and x.ndim == 1):
            x = x[..., None]

        # Generates a distribution of derivatives using the parameter distributions
        if not use_par_dist:
            beta, r_core, norm = self._model_pars
        else:
            beta, r_core, norm = self.par_dists
        return ((2*x)/np.power(r_core, 2))*((-3*beta) + 0.5)*norm*np.power((1+(np.power(x/r_core, 2))), ((-3*beta)-0.5))

    def inverse_abel(self, x: Quantity, use_par_dist: bool = False, method='analytical') -> Quantity:
        """
        This overrides the inverse abel method of the model superclass, as there is an analytical solution to the
        inverse abel transform of the single beta model. The form of the inverse abel transform is that of the
        king profile, but with an extra transformation applied to the normalising parameter. This method can either
        return a single value calculated using the current model parameters, or a distribution of values using
        the parameter distributions (assuming that this model has had a fit run on it).

        :param Quantity x: The x location(s) at which to calculate the value of the inverse abel transform.
        :param bool use_par_dist: Should the parameter distributions be used to calculate a inverse abel transform
            distribution; this can only be used if a fit has been performed using the model instance.
            Default is False, in which case the current parameters will be used to calculate a single value.
        :param str method: The method that should be used to calculate the values of this inverse abel transform.
            Default for this overriding method is 'analytical', in which case the analytical solution is used.
            You  may pass 'direct', 'basex', 'hansenlaw', 'onion_bordas', 'onion_peeling', 'two_point', or
            'three_point' to calculate the transform numerically.
        :return: The inverse abel transform result.
        :rtype: Quantity
        """
        def transform(x_val: Quantity, beta: Quantity, r_core: Quantity, norm: Quantity):
            """
            The function that calculates the inverse abel transform of this beta profile.

            :param Quantity x_val: The x location(s) at which to calculate the value of the inverse abel transform.
            :param Quantity beta: The beta parameter of the beta profile.
            :param Quantity r_core: The core radius parameter of the beta profile.
            :param Quantity norm: The normalisation of the beta profile.
            :return:
            """
            # We calculate the new normalisation parameter
            new_norm = norm / ((gamma((3 * beta) - 0.5) * np.sqrt(np.pi) * r_core) / gamma(3 * beta))

            # Then return the value of the transformed beta profile
            return new_norm * np.power((1 + (np.power(x_val / r_core, 2))), (-3 * beta))

        # Checking x units to make sure that they are valid
        if not x.unit.is_equivalent(self._x_unit):
            raise UnitConversionError("The input x coordinates cannot be converted to units of "
                                      "{}".format(self._x_unit.to_string()))
        else:
            x = x.to(self._x_unit)

        if method == 'analytical':
            # The way the calculation is called depends on whether the user wants to use the parameter distributions
            #  or just the current model parameter values to calculate the inverse abel transform.
            if not use_par_dist:
                transform_res = transform(x, *self.model_pars)
            elif use_par_dist and len(self._par_dists[0]) != 0:
                transform_res = transform(x[..., None], *self.par_dists)
            elif use_par_dist and len(self._par_dists[0]) == 0:
                raise XGAFitError("No fit has been performed with this model, so there are no parameter distributions"
                                  " available.")
        else:
            transform_res = super().inverse_abel(x, use_par_dist, method)

        return transform_res


class DoubleBetaProfile1D(BaseModel1D):
    """
    An XGA model implementation of the double beta profile, a summation of two single beta models. Often thought
    to deal better with peaky cluster cores that you might get from a cool-core cluster, this model can be used
    to describe a galaxy cluster radial surface brightness profile.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('ct/(s*arcmin**2)'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing the surface brightness double-beta
        profile model.
        """

        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('ct/(s*arcmin**2)'), Unit('ct/(s*kpc**2)'), Unit('ct/(s*pix**2)')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        poss_x_units = [kpc, deg, r200, r500, r2500]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        # TODO MAKE THE NEW START PARAMETERS MORE SENSIBLE
        r_core1_starts = [Quantity(100, 'kpc'), Quantity(0.2, 'deg'), Quantity(0.05, r200), Quantity(0.1, r500),
                          Quantity(0.5, r2500)]
        norm_starts = [Quantity(1, 'ct/(s*arcmin**2)'), Quantity(1, 'ct/(s*kpc**2)'), Quantity(1, 'ct/(s*pix**2)')]
        r_core2_starts = [Quantity(400, 'kpc'), Quantity(0.5, 'deg'), Quantity(0.2, r200), Quantity(0.4, r500),
                          Quantity(1, r2500)]

        start_pars = [Quantity(1, ''), r_core1_starts[xu_ind], norm_starts[yu_ind], Quantity(0.5, ''),
                      r_core2_starts[xu_ind], norm_starts[yu_ind]*0.5]
        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        # TODO ALSO MAKE THESE MORE SENSIBLE
        r_core_priors = [{'prior': Quantity([1, 2000], 'kpc'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], 'deg'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r200), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r500), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r2500), 'type': 'uniform'}]
        norm_priors = [{'prior': Quantity([0, 3], 'ct/(s*arcmin**2)'), 'type': 'uniform'},
                       {'prior': Quantity([0, 100], 'ct/(s*kpc**2)'), 'type': 'uniform'},
                       {'prior': Quantity([0, 100], 'ct/(s*pix**2)'), 'type': 'uniform'}]

        priors = [{'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind], norm_priors[yu_ind],
                  {'prior': Quantity([0, 3]), 'type': 'uniform'}, r_core_priors[xu_ind], norm_priors[yu_ind]]

        nice_pars = [r"$\beta_{1}$", r"R$_{\rm{core},1}$", r"S$_{01}$", r"$\beta_{2}$", r"R$_{\rm{core},2}$",
                     r"S$_{02}$"]
        info_dict = {'author': 'placeholder', 'year': 'placeholder', 'reference': 'placeholder',
                     'general': 'The double beta profile, a summation of two single beta models. Often\n '
                                'thought to deal better with peaky cluster cores that you might get from a\n'
                                ' cool-core cluster, this model can be used to describe a galaxy cluster\n'
                                ' radial surface brightness profile.'}
        super().__init__(x_unit, y_unit, start_pars, priors, 'double_beta', 'Double Beta Profile', nice_pars,
                         'Surface Brightness', info_dict)

    @staticmethod
    def model(x: Quantity, beta_one: Quantity, r_core_one: Quantity, norm_one: Quantity, beta_two: Quantity,
              r_core_two: Quantity, norm_two: Quantity) -> Quantity:
        """
        The model function for the double beta profile.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity norm_one: The normalisation of the first beta profile.
        :param Quantity beta_one: The beta slope parameter of the first component beta profile.
        :param Quantity r_core_one: The core radius of the first component beta profile.
        :param Quantity norm_two: The normalisation of the second beta profile.
        :param Quantity beta_two:  The beta slope parameter of the second component beta profile.
        :param Quantity r_core_two: The core radius of the second component beta profile.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """
        p1 = norm_one * ((1 + ((x / r_core_one) ** 2)) ** ((-3 * beta_one) + 0.5))
        p2 = norm_two * ((1 + ((x / r_core_two) ** 2)) ** ((-3 * beta_two) + 0.5))
        return p1 + p2

    def derivative(self, x: Quantity, dx: Quantity = Quantity(0, ''), use_par_dist: bool = False) -> Quantity:
        """
        Calculates the gradient of the double beta profile at a given point, overriding the numerical method
        implemented in the BaseModel1D class, as this simple model has an easily derivable first derivative.

        :param Quantity x: The point(s) at which the slope of the model should be measured.
        :param Quantity dx: This makes no difference here, as this is an analytical derivative. It has
            been left in so that the inputs for this method don't vary between models.
        :param bool use_par_dist: Should the parameter distributions be used to calculate a derivative
            distribution; this can only be used if a fit has been performed using the model instance.
            Default is False, in which case the current parameters will be used to calculate a single value.
        :return: The calculated slope of the model at the supplied x position(s).
        :rtype: Quantity
        """
        # This makes sure that the input radius, or radii, are being used properly. For a single x value, or a set of
        #  x values, we use [..., None] to ensure that (if the user has decided to use the parameter distribution) we
        #  get M distributions of the derivative. In the case where the user has passed a distribution of radii, really
        #  representative of a single radius but with uncertainty, then this does not trigger, and rather than an N
        #  by N (where N is the number of samples in each posterior) array, you get an M x N array
        if x.isscalar or (not x.isscalar and x.ndim == 1):
            x = x[..., None]

        if not use_par_dist:
            beta_one, r_core_one, norm_one, beta_two, r_core_two, norm_two = self._model_pars
        else:
            beta_one, r_core_one, norm_one, beta_two, r_core_two, norm_two = self.par_dists

        p1 = ((2*x)/np.power(r_core_one, 2))*((-3*beta_one) + 0.5)*norm_one*np.power((1+(np.power(x/r_core_one, 2))),
                                                                                     ((-3*beta_one)-0.5))
        p2 = ((2*x)/np.power(r_core_two, 2))*((-3*beta_two)+0.5)*norm_two*np.power((1+(np.power(x/r_core_two, 2))),
                                                                                   ((-3*beta_two)-0.5))
        return p1 + p2

    def inverse_abel(self, x: Quantity, use_par_dist: bool = False, method='analytical') -> Quantity:
        """
        This overrides the inverse abel method of the model superclass, as there is an analytical solution to the
        inverse abel transform of the double beta model. The form of the inverse abel transform is that of two summed
        king profiles, but with extra transformations applied to the normalising parameters. This method can either
        return a single value calculated using the current model parameters, or a distribution of values using
        the parameter distributions (assuming that this model has had a fit run on it).

        :param Quantity x: The x location(s) at which to calculate the value of the inverse abel transform.
        :param bool use_par_dist: Should the parameter distributions be used to calculate a inverse abel transform
            distribution; this can only be used if a fit has been performed using the model instance.
            Default is False, in which case the current parameters will be used to calculate a single value.
        :param str method: The method that should be used to calculate the values of this inverse abel transform.
            Default for this overriding method is 'analytical', in which case the analytical solution is used.
            You  may pass 'direct', 'basex', 'hansenlaw', 'onion_bordas', 'onion_peeling', 'two_point', or
            'three_point' to calculate the transform numerically.
        :return: The inverse abel transform result.
        :rtype: Quantity
        """
        def transform(x_val: Quantity, beta: Quantity, r_core: Quantity, norm: Quantity, beta_two: Quantity,
                      r_core_two: Quantity, norm_two: Quantity):
            """
            The function that calculates the inverse abel transform of this double beta profile.

            :param Quantity x_val: The x location(s) at which to calculate the value of the inverse abel transform.
            :param Quantity beta: The beta parameter of the first beta profile.
            :param Quantity r_core: The core radius parameter of the first beta profile.
            :param Quantity norm: The normalisation of the first beta profile.
            :param Quantity beta_two: The beta parameter of the second beta profile.
            :param Quantity r_core_two: The core radius parameter of the second beta profile.
            :param Quantity norm_two: The normalisation of the second beta profile.
            :return:
            """
            # We calculate the new normalisation parameter
            new_norm = norm / ((gamma((3 * beta) - 0.5) * np.sqrt(np.pi) * r_core) / gamma(3 * beta))
            new_norm_two = norm_two / ((gamma((3 * beta_two) - 0.5) * np.sqrt(np.pi)
                                        * r_core_two) / gamma(3 * beta_two))

            # Then return the value of the transformed beta profile
            return new_norm * np.power((1 + (np.power(x_val / r_core, 2))), (-3 * beta)) + \
                   new_norm_two * np.power((1 + (np.power(x_val / r_core_two, 2))), (-3 * beta_two))

        # Checking x units to make sure that they are valid
        if not x.unit.is_equivalent(self._x_unit):
            raise UnitConversionError("The input x coordinates cannot be converted to units of "
                                      "{}".format(self._x_unit.to_string()))
        else:
            x = x.to(self._x_unit)

        if method == 'analytical':
            # The way the calculation is called depends on whether the user wants to use the parameter distributions
            #  or just the current model parameter values to calculate the inverse abel transform.
            if not use_par_dist:
                transform_res = transform(x, *self.model_pars)
            elif use_par_dist and len(self._par_dists[0]) != 0:
                transform_res = transform(x[..., None], *self.par_dists)
            elif use_par_dist and len(self._par_dists[0]) == 0:
                raise XGAFitError("No fit has been performed with this model, so there are no parameter distributions"
                                  " available.")
        else:
            transform_res = super().inverse_abel(x, use_par_dist, method)

        return transform_res


# So that things like fitting functions can be written generally to support different models
SB_MODELS = {"beta": BetaProfile1D, "double_beta": DoubleBetaProfile1D}
SB_MODELS_PUB_NAMES = {n: m().publication_name for n, m in SB_MODELS.items()}
SB_MODELS_PAR_NAMES = {n: m().par_publication_names for n, m in SB_MODELS.items()}



#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 30/07/2024, 12:44. Copyright (c) The Contributors

from typing import Union, List

from astropy.units import Quantity, Unit, UnitConversionError, kpc

from .base import BaseModel1D


class CoreConstPowerEntropy(BaseModel1D):
    """
    A model to fit galaxy cluster radial entropy profiles, made up of a constant core combined with a power
    law that dominates in the outskirts. The functional form is simple, but was found to be the best model
    for fitting many different entropy profiles by Cavagnolo et al. (https://doi.org/10.1088/0067-0049/182/1/12).

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('keV cm^2'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing a model of how entropy changes with radius -
        this one is almost constant in the core and behaves as a power-law in the outskirts.
        """
        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('keV cm^2')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        # , deg, r200, r500, r2500 - TODO These can't be supported yet because of the normalising 100kpc
        poss_x_units = [kpc]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        k_zero_starts = [Quantity(75, 'keV cm^2')]
        k_100kpc = [Quantity(100, 'keV cm^2')]
        alpha = Quantity(1, '')

        start_pars = [k_zero_starts[yu_ind], k_100kpc[yu_ind], alpha]

        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        kz_priors = [{'prior': Quantity([1, 300], 'keV cm^2'), 'type': 'uniform'}]
        k100_priors = [{'prior': Quantity([1, 500], 'keV cm^2'), 'type': 'uniform'}]
        alpha_priors = {'prior': Quantity([-4, 4], ''), 'type': 'uniform'}

        priors = [kz_priors[yu_ind], k100_priors[yu_ind], alpha_priors]

        nice_pars = [r"K$_{0}$", r"K$_{100\rm{kpc}}$", r"$\alpha$"]
        info_dict = {'author': 'Cavagnolo et al.', 'year': 2009,
                     'reference': 'https://doi.org/10.1088/0067-0049/182/1/12',
                     'general': "A model to fit galaxy cluster radial entropy profiles, made\n"
                                "up of a constant core combined with a power law that\n"
                                "dominates in the outskirts."
                     }

        super().__init__(x_unit, y_unit, start_pars, priors, 'coreconst_power_entropy',
                         'Constant Core and Power-Law Profile', nice_pars, 'Gas Entropy', info_dict)

    @staticmethod
    def model(x: Quantity, k_zero: Quantity, k_100kpc: Quantity, alpha: Quantity) -> Quantity:
        """
        The model function for the constant-core and power-law entropy model.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity k_zero: Parameter quantifying the typical excess of core entropy above the best fitting
            power-law found at larger radii.
        :param Quantity k_100kpc: A normalization for entropy at 100kpc.
        :param Quantity alpha: The power law index.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """
        if type(x) is Quantity:
            result = k_zero + (k_100kpc * (x / Quantity(100, 'kpc'))**alpha)
        else:
            result = k_zero + (k_100kpc * (x / 100)**alpha)

        return result


class PowerEntropy(BaseModel1D):
    """
    A model to fit galaxy cluster radial entropy profiles, made up of a power law. This is another model that was fit
    to a large sample of clusters by Cavagnolo et al. (https://doi.org/10.1088/0067-0049/182/1/12). This is an
    even simpler version of the CoreConstPowerEntropy model.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('keV cm^2'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing a model of how entropy changes with radius -
        this one is almost constant in the core and behaves as a power-law in the outskirts.
        """
        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('keV cm^2')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        # , deg, r200, r500, r2500 - TODO These can't be supported yet because of the normalising 100kpc
        poss_x_units = [kpc]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        k_100kpc = [Quantity(100, 'keV cm^2')]
        alpha = Quantity(1, '')

        start_pars = [k_100kpc[yu_ind], alpha]

        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        k100_priors = [{'prior': Quantity([1, 500], 'keV cm^2'), 'type': 'uniform'}]
        alpha_priors = {'prior': Quantity([-4, 4], ''), 'type': 'uniform'}

        priors = [k100_priors[yu_ind], alpha_priors]

        nice_pars = [r"K$_{100\rm{kpc}}$", r"$\alpha$"]
        info_dict = {'author': 'Cavagnolo et al.', 'year': 2009,
                     'reference': 'https://doi.org/10.1088/0067-0049/182/1/12',
                     'general': "A very simple model to fit galaxy cluster radial\n"
                                "entropy profiles; a power law."
                     }

        super().__init__(x_unit, y_unit, start_pars, priors, 'power_entropy',
                         'Power-Law Profile', nice_pars, 'Gas Entropy', info_dict)

    @staticmethod
    def model(x: Quantity, k_100kpc: Quantity, alpha: Quantity) -> Quantity:
        """
        The model function for the power-law entropy model.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity k_zero: Parameter quantifying the typical excess of core entropy above the best fitting
            power-law found at larger radii.
        :param Quantity k_100kpc: A normalization for entropy at 100kpc.
        :param Quantity alpha: The power law index.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """
        if type(x) is Quantity:
            result = k_100kpc * (x / Quantity(100, 'kpc'))**alpha
        else:
            result = k_100kpc * (x / 100)**alpha

        return result


# So that things like fitting functions can be written generally to support different models
# "power":
ENTROPY_MODELS = {"coreconst_power_entropy": CoreConstPowerEntropy, 'power_entropy': PowerEntropy}
ENTROPY_MODELS_PUB_NAMES = {n: m().publication_name for n, m in ENTROPY_MODELS.items()}
ENTROPY_MODELS_PAR_NAMES = {n: m().par_publication_names for n, m in ENTROPY_MODELS.items()}


#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 20/02/2023, 14:04. Copyright (c) The Contributors


from . import BaseProduct


class EventList(BaseProduct):
    """
    A product class for event lists, largely just used as a means to store information about the event list.

    :param str path: The path to where the event list file SHOULD be located.
    :param str obs_id: The ObsID related to the event list being declared.
    :param str instrument: The instrument related to the event list being declared.
    :param str stdout_str: The stdout from calling the terminal command.
    :param str stderr_str: The stderr from calling the terminal command.
    :param str gen_cmd: The command used to generate the event list.
    """
    def __init__(self, path: str, obs_id: str, instrument: str, stdout_str: str, stderr_str: str,
                 gen_cmd: str):
        """
        The init method of the EventList class.
        """
        super().__init__(path, obs_id, instrument, stdout_str, stderr_str, gen_cmd)
        self._prod_type = "events"









#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 02/08/2024, 10:56. Copyright (c) The Contributors

import os
import warnings
from functools import wraps
# from multiprocessing.dummy import Pool
from multiprocessing import Pool
from subprocess import Popen, PIPE, TimeoutExpired
from typing import Tuple, Union

import fitsio
import pandas as pd
from fitsio import FITS
from tqdm import tqdm

from .. import XSPEC_VERSION
from ..exceptions import XSPECFitError, MultipleMatchError, NoMatchFoundError, XSPECNotFoundError
from ..samples.base import BaseSample
from ..sources import BaseSource


def execute_cmd(x_script: str, out_file: str, src: str, run_type: str, timeout: float) \
        -> Tuple[Union[FITS, str], str, bool, list, list]:
    """
    This function is called for the local compute option. It will run the supplied XSPEC script, then check
    parse the output for errors and check that the expected output file has been created.

    :param str x_script: The path to an XSPEC script to be run.
    :param str out_file: The expected path for the output file of that XSPEC script.
    :param str src: A string representation of the source object that this fit is associated with.
    :param str run_type: A flag that tells this function what type of run this is; e.g. fit or conv_factors.
    :param float timeout: The length of time (in seconds) which the XSPEC script is allowed to run for before being
        killed.
    :return: FITS object of the results, string repr of the source associated with this fit, boolean variable
        describing if this fit can be used, list of any errors found, list of any warnings found.
    :rtype: Tuple[Union[FITS, str], str, bool, list, list]
    """
    if XSPEC_VERSION is None:
        raise XSPECNotFoundError("There is no XSPEC installation detectable on this machine.")

    # We assume the output will be usable to start with
    usable = True

    cmd = "xspec - {}".format(x_script)
    # I add exec to the beginning to make sure that the command inherits the same process ID as the shell, which
    #  allows the timeout to kill the XSPEC run rather than the shell process. Entirely thanks to slayton on
    #   https://stackoverflow.com/questions/4789837/how-to-terminate-a-python-subprocess-launched-with-shell-true
    xspec_proc = Popen("exec " + cmd, shell=True, stdout=PIPE, stderr=PIPE)

    # This makes sure the process is killed if it does timeout
    try:
        out, err = xspec_proc.communicate(timeout=timeout)
    except TimeoutExpired:
        xspec_proc.kill()
        out, err = xspec_proc.communicate()
        # Need to infer the name of the source to supply it in the warning
        source_name = x_script.split('/')[-1].split("_")[0]
        warnings.warn("An XSPEC fit for {} has timed out".format(source_name), stacklevel=2)
        usable = False

    out = out.decode("UTF-8").split("\n")
    err = err.decode("UTF-8").split("\n")

    # We ignore that particular string in the errors identified from stdout because if we don't just it being
    #  present in the if statement in the executed script is enough to make XGA think that the fit failed, even if
    #  that error message was never printed at all
    err_out_lines = [line.split("***Error: ")[-1] for line in out if "***Error" in line
                     if "No acceptable spectra are left after the cleaning step" not in line]
    warn_out_lines = [line.split("***Warning: ")[-1] for line in out if "***Warning" in line]
    err_err_lines = [line.split("***Error: ")[-1] for line in err if "***Error" in line]
    warn_err_lines = [line.split("***Warning: ")[-1] for line in err if "***Warning" in line]

    if usable and len(err_out_lines) == 0 and len(err_err_lines) == 0:
        usable = True
    else:
        usable = False

    error = err_out_lines + err_err_lines
    warn = warn_out_lines + warn_err_lines
    if os.path.exists(out_file + "_info.csv") and run_type == "fit" and usable:
        # The original version of the xga_output.tcl script output everything as one nice neat fits file
        #  but life is full of extraordinary inconveniences and for some reason it didn't work if called from
        #  a Jupyter Notebook. So now I'm going to smoosh all the csv outputs into one fits.
        results = pd.read_csv(out_file + "_results.csv", header="infer")
        # This is the csv with the fit results in, creates new fits file and adds in
        fitsio.write(out_file + ".fits", results.to_records(index=False), extname="results", clobber=True)
        del results

        # The information about individual spectra, exposure times, luminosities etc.
        spec_info = pd.read_csv(out_file + "_info.csv", header="infer")
        # Gets added into the existing file
        fitsio.write(out_file + ".fits", spec_info.to_records(index=False), extname="spec_info")
        del spec_info

        # This finds all of the matching spectrum plot csvs were generated
        rel_path = "/".join(out_file.split('/')[0:-1])
        # This is mostly just used to find how many files there are
        spec_tabs = [rel_path + "/" + sp for sp in os.listdir(rel_path)
                     if "{}_spec".format(out_file) in rel_path + "/" + sp]
        for spec_i in range(1, len(spec_tabs)+1):
            # Loop through and redefine names like this to ensure they're in the right order
            spec_plot = pd.read_csv(out_file + "_spec{}.csv".format(spec_i), header="infer")
            # Adds all the plot tables into the existing fits file in the right order
            fitsio.write(out_file + ".fits", spec_plot.to_records(index=False), extname="plot{}".format(spec_i))
            del spec_plot

        # This reads in the fits we just made
        with FITS(out_file + ".fits") as res_tables:
            tab_names = [tab.get_extname() for tab in res_tables]
            if "results" not in tab_names or "spec_info" not in tab_names:
                usable = False
        # I'm going to try returning the file path as that should be pickleable
        res_tables = out_file + ".fits"
    elif os.path.exists(out_file) and run_type == "conv_factors" and usable:
        res_tables = out_file
        usable = True
    else:
        res_tables = None
        usable = False

    return res_tables, src, usable, error, warn


def xspec_call(xspec_func):
    """
    This is used as a decorator for functions that produce XSPEC scripts. Depending on the
    system that XGA is running on (and whether the user requests parallel execution), the method of
    executing the XSPEC commands will change. This supports multi-threading.
    :return:
    """

    @wraps(xspec_func)
    def wrapper(*args, **kwargs):
        # The first argument of all of these XSPEC functions will be the source object (or a list of),
        # so rather than return them from the XSPEC model function I'll just access them like this.
        if isinstance(args[0], BaseSource):
            sources = [args[0]]
        elif isinstance(args[0], (list, BaseSample)):
            sources = args[0]
        else:
            raise TypeError("Please pass a source object, or a list of source objects.")

        # This is the output from whatever function this is a decorator for
        # First return is a list of paths of XSPEC scripts to execute, second is the expected output paths,
        #  and 3rd is the number of cores to use.
        # run_type describes the type of XSPEC script being run, for instance a fit or a fakeit run to measure
        #  countrate to luminosity conversion constants
        script_list, paths, cores, run_type, src_inds, radii, timeout = xspec_func(*args, **kwargs)
        src_lookup = {repr(src): src_ind for src_ind, src in enumerate(sources)}
        rel_src_repr = [repr(sources[src_ind]) for src_ind in src_inds]

        # Make sure the timeout is converted to seconds, then just stored as a float
        timeout = timeout.to('second').value

        # This is what the returned information from the execute command gets stored in before being parceled out
        #  to source and spectrum objects
        results = {s: [] for s in src_lookup}
        if run_type == "fit":
            desc = "Running XSPEC Fits"
        elif run_type == "conv_factors":
            desc = "Running XSPEC Simulations"

        if len(script_list) > 0:
            # This mode runs the XSPEC locally in a multiprocessing pool.
            with tqdm(total=len(script_list), desc=desc) as fit, Pool(cores) as pool:
                def callback(results_in):
                    """
                    Callback function for the apply_async pool method, gets called when a task finishes
                    and something is returned.
                    """
                    nonlocal fit  # The progress bar will need updating
                    nonlocal results  # The dictionary the command call results are added to
                    if results_in[0] is None:
                        fit.update(1)
                        return
                    else:
                        res_fits, rel_src, successful, err_list, warn_list = results_in
                        results[rel_src].append([res_fits, successful, err_list, warn_list])
                        fit.update(1)

                for s_ind, s in enumerate(script_list):
                    pth = paths[s_ind]
                    src = rel_src_repr[s_ind]
                    pool.apply_async(execute_cmd, args=(s, pth, src, run_type, timeout), callback=callback)
                pool.close()  # No more tasks can be added to the pool
                pool.join()  # Joins the pool, the code will only move on once the pool is empty.

        elif len(script_list) == 0:
            warnings.warn("All XSPEC operations had already been run.", stacklevel=2)

        # Now we assign the fit results to source objects
        for src_repr in results:
            # Made this lookup list earlier, using string representations of source objects.
            # Finds the ind of the list of sources that we should add these results to
            ind = src_lookup[src_repr]
            s = sources[ind]

            # This flag tells this method if the current set of fits are part of an annular spectra or not
            ann_fit = False
            ann_results = {}
            ann_lums = {}
            ann_obs_order = {}

            for res_set in results[src_repr]:
                if len(res_set) != 0 and res_set[1] and run_type == "fit":
                    with FITS(res_set[0]) as res_table:
                        global_results = res_table["RESULTS"][0]
                        model = global_results["MODEL"].strip(" ")

                        # Just define this to check if this is an annular fit or not
                        first_key = res_table["SPEC_INFO"][0]["SPEC_PATH"].strip(" ").split("/")[-1].split('ra')[-1]
                        first_key = first_key.split('_spec.fits')[0]
                        if "_ident" in first_key:
                            ann_fit = True

                        inst_lums = {}
                        obs_order = []
                        for line_ind, line in enumerate(res_table["SPEC_INFO"]):
                            sp_info = line["SPEC_PATH"].strip(" ").split("/")[-1].split("_")
                            # Want to derive the spectra storage key from the file name, this strips off some
                            #  unnecessary info
                            sp_key = line["SPEC_PATH"].strip(" ").split("/")[-1].split('ra')[-1].split('_spec.fits')[0]

                            # If its not an AnnularSpectra fit then we can just fetch the spectrum from the source
                            #  the normal way
                            if not ann_fit:
                                # This adds ra back on, and removes any ident information if it is there
                                sp_key = 'ra' + sp_key
                                # Finds the appropriate matching spectrum object for the current table line
                                spec = s.get_products("spectrum", sp_info[0], sp_info[1], extra_key=sp_key)[0]
                            else:
                                obs_order.append([sp_info[0], sp_info[1]])
                                ann_id = int(sp_key.split("_ident")[-1].split("_")[1])
                                sp_key = 'ra' + sp_key.split('_ident')[0]
                                first_part = sp_key.split('ri')[0]
                                second_part = "_" + "_".join(sp_key.split('ro')[-1].split("_")[1:])

                                ann_sp_key = first_part + "ar" + "_".join(radii[ind].value.astype(str)) + second_part
                                ann_specs = s.get_products("combined_spectrum", extra_key=ann_sp_key)
                                if len(ann_specs) > 1:
                                    raise MultipleMatchError("I have found multiple matches for that AnnularSpectra, "
                                                             "this is the developers fault, not yours.")
                                elif len(ann_specs) == 0:
                                    raise NoMatchFoundError("Somehow I haven't found the AnnularSpectra that you "
                                                            "fitted, this is the developers fault, not yours")
                                else:
                                    ann_spec = ann_specs[0]
                                    spec = ann_spec.get_spectra(ann_id, sp_info[0], sp_info[1])

                            # Adds information from this fit to the spectrum object.
                            spec.add_fit_data(str(model), line, res_table["PLOT"+str(line_ind+1)])

                            # The add_fit_data method formats the luminosities nicely, so we grab them back out
                            #  to help grab the luminosity needed to pass to the source object 'add_fit_data' method
                            processed_lums = spec.get_luminosities(model)
                            if spec.instrument not in inst_lums:
                                inst_lums[spec.instrument] = processed_lums

                        # Ideally the luminosity reported in the source object will be a PN lum, but its not impossible
                        #  that a PN value won't be available. - it shouldn't matter much, lums across the cameras are
                        #  consistent
                        if "pn" in inst_lums:
                            chosen_lums = inst_lums["pn"]
                        # mos2 generally better than mos1, as mos1 has CCD damage after a certain point in its life
                        elif "mos2" in inst_lums:
                            chosen_lums = inst_lums["mos2"]
                        else:
                            chosen_lums = inst_lums["mos1"]

                        if ann_fit:
                            ann_results[spec.annulus_ident] = global_results
                            ann_lums[spec.annulus_ident] = chosen_lums
                            ann_obs_order[spec.annulus_ident] = obs_order

                        elif not ann_fit:
                            # Push global fit results, luminosities etc. into the corresponding source object.
                            s.add_fit_data(model, global_results, chosen_lums, sp_key)

                elif len(res_set) != 0 and res_set[1] and run_type == "conv_factors":
                    res_table = pd.read_csv(res_set[0], dtype={"lo_en": str, "hi_en": str})
                    # Gets the model name from the file name of the output results table
                    model = res_set[0].split("_")[-3]

                    # We can infer the storage key from the name of the results table, just makes it easier to
                    #  grab the correct spectra
                    storage_key = res_set[0].split('/')[-1].split(s.name)[-1][1:].split(model)[0][:-1]

                    # Grabs the ObsID+instrument combinations from the headers of the csv. Makes sure they are unique
                    #  by going to a set (because there will be two columns for each ObsID+Instrument, rate and Lx)
                    # First two columns are skipped because they are energy limits
                    combos = list(set([c.split("_")[1] for c in res_table.columns[2:]]))
                    # Getting the spectra for each column, then assigning rates and lums
                    for comb in combos:
                        spec = s.get_products("spectrum", comb[:10], comb[10:], extra_key=storage_key)[0]
                        spec.add_conv_factors(res_table["lo_en"].values, res_table["hi_en"].values,
                                              res_table["rate_{}".format(comb)].values,
                                              res_table["Lx_{}".format(comb)].values, model)

                elif len(res_set) != 0 and not res_set[1]:
                    for err in res_set[2]:
                        raise XSPECFitError(err)

            if ann_fit:
                # We fetch the annular spectra object that we just fitted, searching by using the set ID of
                #  the last spectra that was opened in the loop
                ann_spec = s.get_annular_spectra(set_id=spec.set_ident)
                try:
                    ann_spec.add_fit_data(model, ann_results, ann_lums, ann_obs_order)

                    # The most likely reason for running XSPEC fits to a profile is to create a temp. profile
                    #  so we check whether constant*tbabs*apec (single_temp_apec function)has been run and if so
                    #  generate a Tx profile automatically
                    if model == "constant*tbabs*apec":
                        temp_prof = ann_spec.generate_profile(model, 'kT', 'keV')
                        s.update_products(temp_prof)

                        # Normalisation profiles can be useful for many things, so we generate them too
                        norm_prof = ann_spec.generate_profile(model, 'norm', 'cm^-5')
                        s.update_products(norm_prof)

                        if 'Abundanc' in ann_spec.get_results(0, 'constant*tbabs*apec'):
                            met_prof = ann_spec.generate_profile(model, 'Abundanc', '')
                            s.update_products(met_prof)

                    else:
                        raise NotImplementedError("How have you even managed to fit this model to a profile?! Its not"
                                                  " supported yet.")
                except ValueError:
                    warnings.warn("{src} annular spectra profile fit was not "
                                  "successful".format(src=ann_spec.src_name), stacklevel=2)

        # If only one source was passed, turn it back into a source object rather than a source
        # object in a list.
        if len(sources) == 1:
            sources = sources[0]
        return sources
    return wrapper






#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 10/11/2023, 14:57. Copyright (c) The Contributors
from datetime import datetime
from typing import Union, List, Tuple
from warnings import warn

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
from astropy.time import Time, TimeDelta
from astropy.units import Quantity, Unit, UnitConversionError
from fitsio import FITS, FITSHDR, read_header
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from xga.exceptions import FailedProductError, IncompatibleProductError, NotAssociatedError
from xga.products import BaseProduct, BaseAggregateProduct
from xga.utils import dict_search


class LightCurve(BaseProduct):
    """
    This is the XGA LightCurve product class, which is used to interface with X-ray lightcurves generated
    for a variety of sources. It provides simple access to data and information about the lightcurve, fitting
    capabilities, and the ability to easily create lightcurve visualisations.

    :param str path: The path to the lightcurve.
    :param str obs_id: The ObsID from which this lightcurve was generated.
    :param str instrument: The instrument from which this lightcurve.
    :param str stdout_str: The stdout from the generation process.
    :param str stderr_str: The stderr for the generation process.
    :param str gen_cmd: The generation command for the lightcurve.
    :param Quantity central_coord: The central coordinate of the region from which this lightcurve was extracted.
    :param Quantity inn_rad: The inner radius of the lightcurve region.
    :param Quantity out_rad: The outer radius of the lightcurve region.
    :param Quantity lo_en: The lower energy bound for this lightcurve.
    :param Quantity hi_en: The upper energy bound for this lightcurve.
    :param Quantity time_bin_size: The time bin size used to generate the lightcurve.
    :param str pattern_expr: The event selection pattern used to generate the lightcurve.
    :param bool region: Whether this was generated from a region in a region file
    :param bool is_back_sub: Whether this lightcurve is background subtracted or not.
    """
    def __init__(self, path: str, obs_id: str, instrument: str, stdout_str: str, stderr_str: str, gen_cmd: str,
                 central_coord: Quantity, inn_rad: Quantity, out_rad: Quantity, lo_en: Quantity, hi_en: Quantity,
                 time_bin_size: Quantity, pattern_expr: str, region: bool = False, is_back_sub: bool = True):
        """
        This is the XGA LightCurve product class, which is used to interface with X-ray lightcurves generated
        for a variety of sources. It provides simple access to data and information about the lightcurve, fitting
        capabilities, and the ability to easily create lightcurve visualisations.

        :param str path: The path to the lightcurve.
        :param str obs_id: The ObsID from which this lightcurve was generated.
        :param str instrument: The instrument from which this lightcurve.
        :param str stdout_str: The stdout from the generation process.
        :param str stderr_str: The stderr for the generation process.
        :param str gen_cmd: The generation command for the lightcurve.
        :param Quantity central_coord: The central coordinate of the region from which this lightcurve was extracted.
        :param Quantity inn_rad: The inner radius of the lightcurve region.
        :param Quantity out_rad: The outer radius of the lightcurve region.
        :param Quantity lo_en: The lower energy bound for this lightcurve.
        :param Quantity hi_en: The upper energy bound for this lightcurve.
        :param Quantity time_bin_size: The time bin size used to generate the lightcurve.
        :param str pattern_expr: The event selection pattern used to generate the lightcurve.
        :param bool region: Whether this was generated from a region in a region file
        :param bool is_back_sub: Whether this lightcurve is background subtracted or not.
        """
        # Unfortunate local import to avoid circular import errors
        from xga.sas import check_pattern

        super().__init__(path, obs_id, instrument, stdout_str, stderr_str, gen_cmd)

        self._prod_type = "lightcurve"

        self._time_bin = time_bin_size
        self._pattern_expr, self._pattern_name = check_pattern(pattern_expr)

        self._energy_bounds = (lo_en, hi_en)

        # Storing the central coordinate of this spectrum
        self._central_coord = central_coord

        # Storing the region information
        self._inner_rad = inn_rad
        self._outer_rad = out_rad
        # And also the shape of the region
        if self._inner_rad.isscalar:
            self._shape = 'circular'
        else:
            self._shape = 'elliptical'

        # This describes whether this spectrum was generated directly from a region present in a region file
        self._region = region

        lc_storage_name = "bound_{l}-{u}_ra{ra}_dec{dec}_ri{ri}_ro{ro}"
        if not self._region and self.inner_rad.isscalar:
            lc_storage_name = lc_storage_name.format(ra=self.central_coord[0].value, dec=self.central_coord[1].value,
                                                     ri=self._inner_rad.value, ro=self._outer_rad.value,
                                                     l=self.energy_bounds[0].to('keV').value,
                                                     u=self.energy_bounds[1].to('keV').value)
        elif not self._region and not self._inner_rad.isscalar:
            inn_rad_str = 'and'.join(self._inner_rad.value.astype(str))
            out_rad_str = 'and'.join(self._outer_rad.value.astype(str))
            lc_storage_name = lc_storage_name.format(ra=self.central_coord[0].value, dec=self.central_coord[1].value,
                                                     ri=inn_rad_str, ro=out_rad_str,
                                                     l=self.energy_bounds[0].to('keV').value,
                                                     u=self.energy_bounds[1].to('keV').value)
        else:
            lc_storage_name = "region"

        # Won't include energy bounds in this right now because I think the update_products method will do that
        #  part for us - bound_{l}-{u}keV l=lo_en.value, u=hi_en.value
        lc_storage_name += "_timebin{tb}_pattern{p}".format(tb=time_bin_size, p=self._pattern_name)

        # And we save the completed key to an attribute
        self._storage_key = lc_storage_name

        # The definition of this product has defined whether the data in 'RATE' are background subtracted or not, and
        #  that info is stored in this attribute - changes how we read in data
        self._is_back_sub = is_back_sub

        # Here we set up attributes to store the various information we can pull from a light curve file - they
        #  are all initially set to None because we only read the information into memory if the user actually
        #  calls one of the properties which uses one of these attributes
        # This will store the background subtracted count rate and uncertainty
        self._bck_sub_cnt_rate = None
        self._bck_sub_cnt_rate_err = None
        # This should store the values (in counts/per second) for the source, instrumental-effect-corrected, count-
        #  rate of the light curve
        self._src_cnt_rate = None
        # The count-rate should be accompanied by uncertainties, and they will live in this attribute
        self._src_cnt_rate_err = None
        # This stores an important quantity, the reference time for the light curve time values
        self._ref_time = None
        # This stores the 'time system' - so for default XMM lightcurves for instance it will be TT (terrestrial time)
        self._time_sys = None
        # Here we store the x-axis, the time steps which the count-rates are attributed to
        self._time = None
        # The fractional exposure time of the livetime for each data point
        self._frac_exp = None
        # The background count-rate and count-rate uncertainty, scaled to account for the area difference between
        #  the source and background regions
        self._bck_cnt_rate = None
        self._bck_cnt_rate_err = None

        # These attributes store 2D quantities that describe the start and end points of the good-time-intervals for
        #  the source and background light-curves. The zeroth column is start, and the other column is stop
        self._src_gti = None
        self._bck_gti = None

        # In these attributes we store the XMM recorded start and stop times for this light curve, as well as the
        #  place where the time values were assigned (i.e. relative to satellite or barycentre of the solar system)
        self._time_start = None
        self._time_stop = None
        self._time_assign = None

        # This just stores whether the data have been read into memory or not (set by the _read_on_demand method)
        self._read_in = False

        # This is used to store the header of the main data table, but again it is only read in its entirety if the
        #  user actually asks for it through the 'header' property
        self._header = None

    # Defining properties first
    @property
    def storage_key(self) -> str:
        """
        This property returns the storage key which this object assembles to place the LightCurve in
        an XGA source's storage structure. The key is based on the properties of the LightCurve, and
        some configuration options, and is basically human-readable.

        :return: String storage key.
        :rtype: str
        """
        return self._storage_key

    @property
    def central_coord(self) -> Quantity:
        """
        This property provides the central coordinates (RA-Dec) of the region that this light curve
        was generated from.

        :return: Astropy Quantity object containing the central coordinate in degrees.
        :rtype: Quantity
        """
        return self._central_coord

    @property
    def shape(self) -> str:
        """
        Returns the shape of the outer edge of the region this light curve was generated from.

        :return: The shape (either circular or elliptical).
        :rtype: str
        """
        return self._shape

    @property
    def inner_rad(self) -> Quantity:
        """
        Gives the inner radius (if circular) or radii (if elliptical - semi-major, semi-minor) of the
        region in which this light curve was generated.

        :return: The inner radius(ii) of the region.
        :rtype: Quantity
        """
        return self._inner_rad

    @property
    def outer_rad(self) -> Quantity:
        """
        Gives the outer radius (if circular) or radii (if elliptical - semi-major, semi-minor) of the
        region in which this light curve was generated.

        :return: The outer radius(ii) of the region.
        :rtype: Quantity
        """
        return self._outer_rad

    @property
    def time_bin_size(self) -> Quantity:
        """
        Gives the time bin size used to generate the lightcurve.

        :return: The time bin size used to generate the lightcurve.
        :rtype: Quantity
        """
        return self._time_bin

    @property
    def count_rate(self) -> Quantity:
        """
        Returns the background-subtracted instrumental-effect-corrected count-rates for this light curve.

        :return: Background-subtracted instrumental-effect-corrected count-rates, in units of ct/s.
        :rtype: Quantity
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return self._bck_sub_cnt_rate

    @property
    def count_rate_err(self) -> Quantity:
        """
        Returns the background-subtracted instrumental-effect-corrected count-rate uncertainties for this light curve.

        :return: Background-subtracted instrumental-effect-corrected count-rate uncertainties, in units of ct/s.
        :rtype: Quantity
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return self._bck_sub_cnt_rate_err

    @property
    def src_count_rate(self) -> Quantity:
        """
        Returns the source instrumental-effect-corrected count-rates for this light curve.

        :return: Source instrumental-effect-corrected count-rates, in units of ct/s.
        :rtype: Quantity
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return self._src_cnt_rate

    @property
    def src_count_rate_err(self) -> Quantity:
        """
        Returns the source instrumental-effect-corrected count-rate uncertainties for this light curve.

        :return: Source instrumental-effect-corrected count-rate uncertainties, in units of ct/s.
        :rtype: Quantity
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return self._src_cnt_rate_err

    @property
    def ref_time(self) -> Time:
        """
        Returns the reference time for this lightcurve, which is what the 'time' values are calculated from.

        :return: An Astropy Time object that defines the reference time for this lightcurve.
        :rtype: Time
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return self._ref_time

    @property
    def time_system(self) -> str:
        """
        Returns the time system for this lightcurve; e.g. TT or terrestrial time.

        :return: The time system.
        :rtype: str
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return self._time_sys

    @property
    def time(self) -> Quantity:
        """
        Returns the time steps that correspond to the count-rates measured for this light curve

        :return: Background-subtracted and instrumental-effect-corrected count-rate uncertainties, in units of seconds.
        :rtype: Quantity
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return self._time

    @property
    def datetime(self) -> datetime:
        """
        Returns the time steps for this light curve, but in a datetime format, and no longer relative to a
        reference time.

        :return: The absolute datetimes that the time steps correspond to.
        :rtype: np.ndarray(datetime)
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return (self.ref_time + TimeDelta(self.time, format='sec', scale=self.time_system.lower())).to_datetime()

    @property
    def bck_count_rate(self) -> Quantity:
        """
        Returns the background count-rates for this light curve.

        :return: Background count-rates, in units of ct/s.
        :rtype: Quantity
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return self._bck_cnt_rate

    @property
    def bck_count_rate_err(self) -> Quantity:
        """
        Returns the background count-rate uncertainties for this light curve.

        :return: Background count-rate uncertainties, in units of ct/s.
        :rtype: Quantity
        """
        # The version of read on demand for this class will itself check to see if the data have already been
        #  read in, so we don't need to check that here - this method will read our LC data from disk into this
        #  XGA product instance
        self._read_on_demand()

        return self._bck_cnt_rate_err

    @property
    def frac_exp(self) -> Quantity:
        """
        The fractional exposure time for each entry in this light curve.

        :return: Fractional exposure.
        :rtype: Quantity
        """
        # Make sure the file is read in
        self._read_on_demand()
        return self._frac_exp

    @property
    def src_gti(self) -> Quantity:
        """
        Returns a 2D quantity with start (column 0) and end (column 1) times for the good-time-intervals of the
        source light curve.

        :return: A 2D astropy quantity with start (column 0) and end (column 1) times for the source
            good-time-intervals (in seconds).
        :rtype: Quantity
        """
        # Make sure the GTI information is read into memory
        self._read_on_demand()
        return self._src_gti

    @property
    def bck_gti(self) -> Quantity:
        """
        Returns a 2D quantity with start (column 0) and end (column 1) times for the good-time-intervals of the
        background light curve.

        :return: A 2D astropy quantity with start (column 0) and end (column 1) times for the background
            good-time-intervals (in seconds).
        :rtype: Quantity
        """
        # Make sure the GTI information is read into memory
        self._read_on_demand()
        return self._bck_gti

    @property
    def start_time(self) -> Quantity:
        """
        A property getter to access the recorded start time for this light curve.

        :return: Light curve start time, in seconds.
        :rtype: Quantity
        """
        self._read_on_demand()
        return self._time_start

    @property
    def stop_time(self) -> Quantity:
        """
        A property getter to access the recorded stop time for this light curve.

        :return: Light curve stop time, in seconds.
        :rtype: Quantity
        """
        self._read_on_demand()
        return self._time_stop

    @property
    def start_datetime(self) -> datetime:
        """
        A property getter to access the recorded start datetime for this light curve.

        :return: Light curve start datetime.
        :rtype: datetime
        """
        self._read_on_demand()

        return (self.ref_time + TimeDelta(self._time_start, format='sec', scale=self.time_system.lower())).to_datetime()

    @property
    def stop_datetime(self) -> datetime:
        """
        A property getter to access the recorded stop datetime for this light curve.

        :return: Light curve stop datetime.
        :rtype: datetime
        """
        self._read_on_demand()
        return (self.ref_time + TimeDelta(self._time_stop, format='sec', scale=self.time_system.lower())).to_datetime()

    @property
    def time_assign(self) -> str:
        """
        A property getter to access the physical location that the assigned times are based on.

        :return: The TASSIGN entry of the light curve file.
        :rtype: str
        """
        self._read_on_demand()
        return self._time_assign

    @property
    def header(self) -> FITSHDR:
        """
        Property getter allowing access to the main fits header of the light curve.

        :return: The header of the primary data table (RATE) of the light curve that was read in.
        :rtype: FITSHDR
        """
        if self._header is None and self.usable:
            self._header = read_header(self.path)
        elif not self.usable:
            reasons = ", ".join(self.not_usable_reasons)
            raise FailedProductError("SAS failed to generate this product successfully, so you cannot access "
                                     "the header from it; reason given is {}.".format(reasons))

        return self._header

    # Then define internal methods
    def _read_on_demand(self):
        """
        This method is called by properties that deliver data to the user, either directly or via other methods of
        this class, such as view(). It will ensure that the data haven't already been read from the source file into
        memory, and that the source file has actually been classed as usable, and then read the relevant data into
        attributes of this class.
        """
        # TODO The way this is laid out is a bit weird

        # Usable flag to check that nothing went wrong in the light-curve generation, and the _read_in flag to
        #  check that we haven't already read this in to memory - no sense doing it again
        if self.usable and not self._read_in:
            with FITS(self.path) as all_lc:
                # This chunk reads out the various columns of the 'RATE' entry in the light curve file, storing
                #  them in suitably unit-ed astropy quantities
                if self._is_back_sub:
                    # TODO I should email the XMM help desk about this and double check
                    self._bck_sub_cnt_rate = Quantity(all_lc['RATE'].read_column('RATE'), 'ct/s')
                    self._bck_sub_cnt_rate_err = Quantity(all_lc['RATE'].read_column('ERROR'), 'ct/s')

                    if np.isnan(self._bck_sub_cnt_rate).any():
                        good_ent = np.where(~np.isnan(self._bck_sub_cnt_rate))
                        self._bck_sub_cnt_rate = self._bck_sub_cnt_rate[good_ent]
                        self._bck_sub_cnt_rate_err = self._bck_sub_cnt_rate_err[good_ent]
                    else:
                        good_ent = np.arange(0, len(self._bck_sub_cnt_rate))
                else:
                    # If we weren't told that the rate data are background subtracted when the light curve was
                    #  declared, then we store the values in the source count rate attributes
                    self._src_cnt_rate = Quantity(all_lc['RATE'].read_column('RATE'), 'ct/s')
                    self._src_cnt_rate_err = Quantity(all_lc['RATE'].read_column('ERROR'), 'ct/s')

                    if np.isnan(self._src_cnt_rate).any():
                        good_ent = np.where(~np.isnan(self._src_cnt_rate))
                        self._src_cnt_rate = self._src_cnt_rate[good_ent]
                        self._src_cnt_rate_err = self._src_cnt_rate_err[good_ent]
                    else:
                        good_ent = np.arange(0, len(self._src_cnt_rate))

                self._time = Quantity(all_lc['RATE'].read_column('TIME'), 's')[good_ent]
                self._frac_exp = Quantity(all_lc['RATE'].read_column('FRACEXP'))[good_ent]
                self._bck_cnt_rate = Quantity(all_lc['RATE'].read_column('BACKV'), 'ct/s')[good_ent]
                self._bck_cnt_rate_err = Quantity(all_lc['RATE'].read_column('BACKE'), 'ct/s')[good_ent]

                # Here we read out the beginning and end times of the GTIs for source and background
                self._src_gti = Quantity([all_lc['SRC_GTIS'].read_column('START'),
                                          all_lc['SRC_GTIS'].read_column('STOP')], 's').T
                self._bck_gti = Quantity([all_lc['BKG_GTIS'].read_column('START'),
                                          all_lc['BKG_GTIS'].read_column('STOP')], 's').T

                # Grab the start, stop, and time assign values from the overall header of the light curve
                hdr = all_lc['RATE'].read_header()
                self._time_start = Quantity(hdr['TSTART'], 's')
                self._time_stop = Quantity(hdr['TSTOP'], 's')
                self._time_assign = hdr['TASSIGN']
                self._ref_time = Time(hdr['MJDREF'], format='mjd')
                self._time_sys = hdr['TIMESYS']

            # TODO add calculation for error prop of src-bck or bck+bckcorr
            # And set this attribute to make sure that no further reading in is done
            self._read_in = True

        elif not self.usable:
            reasons = ", ".join(self.not_usable_reasons)
            raise FailedProductError("SAS failed to generate this product successfully, so you cannot access "
                                     "data from it; reason given is {}.".format(reasons))

    def overlap_check(self, lightcurves: Union['LightCurve', List['LightCurve']]) -> Union[np.ndarray, bool]:
        """
        A simple method which checks whether a passed LightCurve (or list of lightcurves) overlap temporally with
        this lightcurve.

        :param LightCurve/List[LightCurve] lightcurves: A LightCurve, or a list of LightCurves, to check for overlap
            with this LightCurve.
        :rtype: np.ndarray/bool
        :return: A boolean value (or an array of boolean values if multiple LightCurve instances were passed) which
            is True if the passed LightCurve temporally overlaps with this one, and False if it does not.
        """
        # Make sure that the input is iterable to normalise the behaviour later on
        if isinstance(lightcurves, LightCurve):
            lightcurves = [lightcurves]

        # Grabs the start and stop times for the passed lightcurves, puts them all in non-scalar quantities
        starts = Quantity([lc.start_time for lc in lightcurves])
        ends = Quantity([lc.stop_time for lc in lightcurves])

        # Simply constructs a boolean array that tells us if each lightcurve starts in, finishes in, or completely
        #  encloses the light curve we're checking against
        overlap = ((starts >= self.start_time) & (starts < self.stop_time)) | \
                  ((ends >= self.start_time) & (ends < self.stop_time)) | \
                  ((starts <= self.start_time) & (ends >= self.stop_time))

        if len(overlap) == 1:
            overlap = overlap[0]

        return overlap

    # Then define user-facing methods
    def get_view(self, ax: Axes, time_unit: Union[str, Unit] = Unit('s'), lo_time_lim: Quantity = None,
                 hi_time_lim: Quantity = None, colour: str = 'black', plot_sep: bool = False,
                 src_colour: str = 'tab:cyan', bck_colour: str = 'firebrick', custom_title: str = None,
                 label_font_size: int = 15, title_font_size: int = 18, highlight_bad_times: bool = True):

        if isinstance(time_unit, str):
            time_unit = Unit(time_unit)

        if not self.time.unit.is_equivalent(time_unit):
            raise UnitConversionError("You have supplied a 'time_unit' that cannot be converted to seconds.")

        time_x = self.time.to(time_unit) - self.start_time.to(time_unit)

        if lo_time_lim is None:
            lo_time_lim = time_x.min()
        elif lo_time_lim is not None and lo_time_lim.unit.is_equivalent(time_unit):
            lo_time_lim = lo_time_lim.to(time_unit)

        if hi_time_lim is None:
            hi_time_lim = time_x.max()
        elif hi_time_lim is not None and hi_time_lim.unit.is_equivalent(time_unit):
            hi_time_lim = hi_time_lim.to(time_unit)

        if plot_sep:
            if self.src_count_rate is None:
                raise ValueError("This light-curve is background subtracted, so we cannot plot the total and "
                                 "background separately.")
            ax.errorbar(time_x.value, self.src_count_rate.value, yerr=self.src_count_rate_err.value, capsize=2,
                        color=src_colour, label='Source', fmt='x')
            ax.errorbar(time_x.value, self.bck_count_rate.value, yerr=self.bck_count_rate_err.value, capsize=2,
                        color=bck_colour, label='Background', fmt='x')
        else:
            ax.errorbar(time_x.value, self.count_rate.value, yerr=self.count_rate_err.value, capsize=2,
                        color=colour, label='Background subtracted', fmt='x')

        if highlight_bad_times and (len(self.src_gti) != 1 or self.src_gti[0, 0] != self.start_time
                                    or self.src_gti[0, 1] != self.stop_time):
            for ind in range(len(self.src_gti)):
                if ind == 0:
                    # This is where the first GTI DOESN'T begin with the start of the observation - so there is a
                    #  bad period between the start of the LC time coverage and the first GTI start
                    if (self.src_gti[ind, 0] - self.start_time).to('s') != 0:
                        bad_start = Quantity(0, 's')
                    else:
                        bad_start = self.src_gti[ind, 1] - self.start_time.to(time_unit)

                    # Here the first BAD time interval goes from the end of the first GTI to the start of the
                    #  next one, OR to the end of the obs if there isn't a next one
                    if len(self.src_gti) != 1:
                        bad_stop = self.src_gti[ind + 1, 0] - self.start_time.to(time_unit)
                    else:
                        bad_stop = self.stop_time - self.start_time

                    # bad_stop = self.src_gti[ind, 0] - self.start_time.to(time_unit)
                    label = "Bad time interval"
                elif ind == (len(self.src_gti)-1):
                    bad_start = self.src_gti[ind, 1]
                    bad_stop = self.stop_time - self.start_time
                    label = ""
                else:
                    bad_start = self.src_gti[ind, 1] - self.start_time.to(time_unit)
                    bad_stop = self.src_gti[ind+1, 0] - self.start_time.to(time_unit)
                    label = ""

                ax.axvspan(bad_start.value, bad_stop.value, color='firebrick', alpha=0.3, label=label)

        if custom_title is not None:
            ax.set_title(custom_title, fontsize=title_font_size)
        elif self.src_name is not None:
            ax.set_title("{s} {t} {o} {i} {l}-{u}keV Lightcurve".format(s=self.src_name, t='XMM', o=self.obs_id,
                                                                        i=self.instrument.upper(),
                                                                        l=self.energy_bounds[0].to('keV').value,
                                                                        u=self.energy_bounds[1].to('keV').value),
                         fontsize=title_font_size)
        else:
            ax.set_title("{t} {o} {i} {l}-{u}keV Lightcurve".format(s=self.src_name, t='XMM', o=self.obs_id,
                                                                    i=self.instrument.upper(),
                                                                    l=self.energy_bounds[0].to('keV').value,
                                                                    u=self.energy_bounds[1].to('keV').value),
                         fontsize=title_font_size)

        if lo_time_lim < time_x.min():
            warn('The lower time limit is smaller than the lowest time value, it has been set to the '
                 'lowest available value.', stacklevel=2)
            lo_time_lim = time_x.min()
        if hi_time_lim > time_x.max():
            warn('The upper time limit is higher than the greatest time value, it has been set to the '
                 'greatest available value.', stacklevel=2)
            hi_time_lim = time_x.max()

        ax.minorticks_on()
        ax.tick_params(direction='in', which='both', right=True, top=True)

        # Setting the axis limits
        ax.set_xlim(lo_time_lim.value, hi_time_lim.value)

        ax.set_xlabel("Relative Time [{}]".format(time_unit.to_string('latex')), fontsize=label_font_size)
        ax.set_ylabel("Count-rate [{}]".format(self.count_rate.unit.to_string('latex')), fontsize=label_font_size)
        ax.legend(loc='best')

        return ax

    def view(self, figsize: tuple = (14, 6), time_unit: Union[str, Unit] = Unit('s'),
             lo_time_lim: Quantity = None, hi_time_lim: Quantity = None, colour: str = 'black',
             plot_sep: bool = False, src_colour: str = 'tab:cyan', bck_colour: str = 'firebrick',
             custom_title: str = None, label_font_size: int = 15, title_font_size: int = 18,
             highlight_bad_times: bool = True):

        # Create figure object
        fig = plt.figure(figsize=figsize)

        ax = plt.gca()

        ax = self.get_view(ax, time_unit, lo_time_lim, hi_time_lim, colour, plot_sep, src_colour, bck_colour,
                           custom_title, label_font_size, title_font_size, highlight_bad_times)
        plt.tight_layout()
        # Display the image
        plt.show()

        # Wipe the figure
        plt.close("all")


class AggregateLightCurve(BaseAggregateProduct):
    """
    The init method for the AggregateLightCurve class, performs checks and organises the light-curves which
    have been passed in, for easy retrieval. It also allows for analysis to be performed on the combined
    data, and for visualisations to be created.

    This class is designed to package light-curves generated for the same source, with the same settings, and
    for the same energy bounds - if interested in the time varying behaviours of multiple energy bands then
    the HardnessCurve and AggregateHardnessCurve products should be used. It can take light-curves from different
    instruments, and will deal with them simultaneously rather than stacking them.

    Light curves that are part of an AggregateLightCurve will be separated into 'time chunks', where a time chunk
    is a period that has uninterrupted coverage. For instance, three XMM observations separated by a year each would
    be in three different time chunks, but if there were a fourth observation that was taken by another telescope
    and happened concurrently (even if it didn't start and end at the same time) with the first XMM
    observation, then it would be in the same time chunk.

    :param Union[List[LightCurve], np.ndarray] lightcurves: A list or array of LightCurve objects that are to be
        collated in an AggregateLightCurve. These must be for the same source, and generated with the same
        settings.
    """
    def __init__(self, lightcurves: Union[List[LightCurve], np.ndarray]):
        """
        The init method for the AggregateLightCurve class, performs checks and organises the light-curves which
        have been passed in, for easy retrieval. It also allows for analysis to be performed on the combined
        data, and for visualisations to be created.

        This class is designed to package light-curves generated for the same source, with the same settings, and
        for the same energy bounds - if interested in the time varying behaviours of multiple energy bands then
        the HardnessCurve and AggregateHardnessCurve products should be used. It can take light-curves from different
        instruments, and will deal with them simultaneously rather than stacking them.

        Light curves that are part of an AggregateLightCurve will be separated into 'time chunks', where a time chunk
        is a period that has uninterrupted coverage. For instance, three XMM observations separated by a year each would
        be in three different time chunks, but if there were a fourth observation that was taken by another telescope
        and happened concurrently (even if it didn't start and end at the same time) with the first XMM
        observation, then it would be in the same time chunk.

        :param Union[List[LightCurve], np.ndarray] lightcurves: A list or array of LightCurve objects that are to be
            collated in an AggregateLightCurve. These must be for the same source, and generated with the same
            settings.
        """
        if isinstance(lightcurves, list):
            lightcurves = np.array(lightcurves)
        elif not isinstance(lightcurves, np.ndarray):
            raise TypeError("The AggregateLightCurve argument must be either a list of LightCurve objects, or a "
                            "numpy array of LightCurve Objects.")

        # Check that all the elements of the list/array are actually LightCurve objects
        if any([not isinstance(lc, LightCurve) for lc in lightcurves]):
            raise TypeError("You cannot pass a list/array containing an object that is not an instance of LightCurve.")

        # Obviously we need to make sure there are enough light curves to aggregate
        if len(lightcurves) < 2:
            raise ValueError("At least two light curves must be passed in order to declare an AggregateLightCurve.")

        # We need to do some checking on the input light-curves, to make sure they're for the same source, region, and
        #  energy bands - the first of the light curves is selected as the comparison point. We're going to check that
        #  each of them has the same central coordinates, region shape, inner radius, outer radius, and energy bounds
        comp_lc = lightcurves[0]
        if not all([np.array_equal(comp_lc.central_coord, lc.central_coord) for lc in lightcurves[1:]]):
            raise IncompatibleProductError("Central coordinates of all lightcurves passed to AggregateLightCurve must "
                                           "be the same.")
        elif not all([lc.shape == comp_lc.shape for lc in lightcurves[1:]]):
            raise IncompatibleProductError("Region shape of all lightcurves passed to AggregateLightCurve must "
                                           "be the same.")
        elif not all([np.array_equal(comp_lc.inner_rad, lc.inner_rad) for lc in lightcurves[1:]]):
            raise IncompatibleProductError("Inner radii of lightcurves passed to AggregateLightCurve must "
                                           "be the same.")
        elif not all([np.array_equal(comp_lc.outer_rad, lc.outer_rad) for lc in lightcurves[1:]]):
            raise IncompatibleProductError("Outer radii of lightcurves passed to AggregateLightCurve must "
                                           "be the same.")
        # Past me made the energy_bounds return a tuple for some reason - just need to convert to a Quantity to compare
        elif not all([np.array_equal(Quantity(comp_lc.energy_bounds), Quantity(lc.energy_bounds))
                      for lc in lightcurves[1:]]):
            raise IncompatibleProductError("The energy bounds of lightcurves passed to AggregateLightCurve must "
                                           "be the same - if interested in the time varying behaviours of multiple "
                                           "energy bands then the HardnessCurve and AggregateHardnessCurve products "
                                           "should be used.")
        # Checks that the time bin sizes used to generate the lightcurves are the same
        elif not all([lc.time_bin_size == comp_lc.time_bin_size for lc in lightcurves[1:]]):
            raise IncompatibleProductError("Time bin sizes of lightcurves passed to AggregateLightCurve must "
                                           "be the same.")
        elif len(list(set([lc.src_name for lc in lightcurves]))) > 1:
            raise ValueError("Some of the passed light curves were not generated for the same object; their "
                             "source names must match.")

        # Pulls out all the ObsIDs and instruments associated with these light curves
        obs_ids = list(set([lc.obs_id for lc in lightcurves]))
        insts = list(set([lc.instrument for lc in lightcurves]))

        if len(obs_ids) == 1:
            obs_id_to_pass = obs_ids[0]
        else:
            obs_id_to_pass = 'combined'

        if len(insts) == 1:
            inst_to_pass = insts[0]
        else:
            inst_to_pass = 'combined'

        # This just sorts the lightcurves by their start time, for earliest to latest
        start_sort = np.argsort(Quantity([lc.start_time for lc in lightcurves]))
        lightcurves = lightcurves[start_sort]

        super().__init__([lc.path for lc in lightcurves], 'lightcurve', obs_id_to_pass, inst_to_pass)

        # Storing the name of the object for which we've generated light curves - we've already checked that all the
        #  input light curves have the same src name.
        self._src_name = lightcurves[0].src_name

        # Make sure to add the energy bounds to the attribute - we require and know that they are the same for
        #  all the lightcurves that have been passed.
        self._energy_bounds = lightcurves[0].energy_bounds

        # TODO This will need some TLC when support for multiple telescopes is implemented
        # This sets the storage key as the same as the LightCurve, but we do account for different patterns accepted
        #  for different instruments - as mos1 and 2 should be treated the same we don't look at the specific MOS
        #  instrument. Having particular instrument behaviour like that will have to be re-examined for
        #  non-XMM lightcurves
        self._patterns = {}
        for lc in lightcurves:
            patt = lc.storage_key.split('_pattern')[-1]
            # Turns mos1 and mos2 into just mos
            rel_inst = lc.instrument.replace('1', '').replace('2', '')
            if rel_inst not in self._patterns:
                self._patterns[rel_inst] = patt
            elif lc.instrument and self._patterns[rel_inst] != patt:
                raise IncompatibleProductError("Lightcurves for the same instrument ({}) must have the same event "
                                               "selection pattern.".format(rel_inst.upper()))

        patts = [pk + 'pattern' + pv for pk, pv in self._patterns.items()]
        # This is what the AggregateLightCurve will be stored under in an XGA source product storage structure.
        self._storage_key = lightcurves[0].storage_key.split('_pattern')[0] + '_' + '_'.join(patts)

        # This stores the ObsIDs and their instruments, for the lightcurves associated with this objcet
        self._rel_obs = {}
        # This array determines which lightcurves have overlapping temporal coverage
        overlapping = np.full((len(lightcurves), len(lightcurves)), False)
        for lc_ind, lc in enumerate(lightcurves):
            # If an ObsID is already present in the rel_obs dict then we need to add the current instrument as part
            #  of a new list, otherwise we need to append our current instrument to an existing list
            if lc.obs_id not in self._rel_obs:
                self._rel_obs[lc.obs_id] = [lc.instrument]
            else:
                self._rel_obs[lc.obs_id].append(lc.instrument)

            # Use the LightCurve overlap checking method to figure out which of our set of light curves overlaps with
            #  the current one. Checking like this does include the current light curve which obviously will overlap,
            #  but we're setting up an overlap matrix and those values will be the diagonal, so we don't mind
            cur_overlap = lc.overlap_check(lightcurves)
            overlapping[lc_ind, :] = cur_overlap

        # Get the entries one up from the diagonal, we'll use them to split the light curves up into time chunks, which
        #  is where there are gaps between coverage. This works because there will be False overlap entries in the
        #  shifted diagonal of the matrix, and those are cases where there is a break in the observations
        split = np.diag(overlapping, 1)
        split = np.insert(split, 0, False)

        # Here we want to index the light curves into time chunk groupings
        groupings = np.zeros(len(lightcurves), dtype=int)
        # This gives us the indices of where we need to split the light curves into chunks, as we know that the light
        #  curves are already sorted into the correct temporal order
        split_inds = np.argwhere(~split).T[0]
        # Just sets up the time chunk ID 1D matrix for our light curves
        for split_ind_ind, split_ind in enumerate(split_inds):
            if split_ind_ind != (len(split_inds)-1):
                groupings[split_ind: split_inds[split_ind_ind+1]] = split_ind_ind
            else:
                # As I want it to go to the very end of the array
                groupings[split_ind:] = split_ind_ind

        self._time_chunk_ids = np.arange(0, len(split_inds))

        # Maybe there is a more elegant, in-line, way of doing this, but I cannot be bothered to think of it
        for lc_ind, lc in enumerate(lightcurves):
            rel_grp = groupings[lc_ind]
            if lc.obs_id not in self._component_products:
                self._component_products[lc.obs_id] = {lc.instrument: {rel_grp: lc}}
            elif lc.obs_id in self._component_products and lc.instrument not in self._component_products[lc.obs_id]:
                self._component_products[lc.obs_id][lc.instrument] = {rel_grp: lc}
            elif lc.obs_id in self._component_products and lc.instrument in self._component_products[lc.obs_id]:
                self._component_products[lc.obs_id][lc.instrument][rel_grp] = lc

    # Start by defining properties, then internal (protected) methods, and then user facing methods
    @property
    def obs_ids(self) -> list:
        """
        A property of this spectrum set that details which ObsIDs have contributed lightcurves to this object.

        :return: A list of ObsIDs.
        :rtype: list
        """
        return list(self._rel_obs.keys())

    @property
    def instruments(self) -> dict:
        """
        A property of this spectrum set that details which ObsIDs and instruments have contributed lightcurves
        to this object. The top level keys are ObsIDs, and the values are lists of instruments.

        :return: A dictionary of lists, with the top level keys being ObsIDs, and the lists
            containing instruments associated with those ObsIDs.
        :rtype: dict
        """
        return self._rel_obs

    @property
    def src_name(self) -> str:
        """
        Method to return the name of the object a product is associated with. The product becomes
        aware of this once it is added to a source object.

        :return: The name of the source object this product is associated with.
        :rtype: str
        """
        return self._src_name

    # This needs a setter, as this property only becomes not-None when the product is added to a source object.
    @src_name.setter
    def src_name(self, name: str):
        """
        Property setter for the src_name attribute of a product, should only really be called by a source object,
        not by a user.

        :param str name: The name of the source object associated with this product.
        """
        self._src_name = name
        for p in self.all_lightcurves:
            p.src_name = name

    @property
    def all_lightcurves(self) -> List[LightCurve]:
        """
        Simple extra wrapper for get_lightcurve that allows the user to retrieve every single lightcurve associated
        with this AggregateLightCurve instance, for all time chunk IDs, ObsIDs, and Instruments.

        :return: A list of every single lightcurve associated with this object.
        :rtype: List[LightCurve]
        """
        # This stores the flattened set of all lightcurves
        all_lcs = []
        # Iterate through the time chunks associated with this AggregateLightCurve, as we need to pass an ID to
        #  get_lightcurves to retrieve lightcurves
        for tc_i in self.time_chunk_ids:
            # If there is only one light curve per time chunk then get_lightcurves will just return a
            #  LightCurve object, but I always want it to be a list
            rel_lcs = self.get_lightcurves(tc_i)
            if isinstance(rel_lcs, LightCurve):
                rel_lcs = [rel_lcs]

            # Add the list of lightcurves for the current time chunk to the overall flattened lightcurve list
            all_lcs += rel_lcs

        return all_lcs

    @property
    def central_coord(self) -> Quantity:
        """
        This property provides the central coordinates (RA-Dec) of the region that this set of light curves
        was generated from.

        :return: Astropy Quantity object containing the central coordinate in degrees.
        :rtype: Quantity
        """
        return self.all_lightcurves[0].central_coord

    @property
    def shape(self) -> str:
        """
        Returns the shape of the outer edge of the region this set of light curves was generated from.

        :return: The shape (either circular or elliptical).
        :rtype: str
        """
        return self.all_lightcurves[0].shape

    @property
    def inner_rad(self) -> Quantity:
        """
        Gives the inner radius (if circular) or radii (if elliptical - semi-major, semi-minor) of the
        region in which this set of light curves was generated.

        :return: The inner radius(ii) of the region.
        :rtype: Quantity
        """
        return self.all_lightcurves[0].inner_rad

    @property
    def outer_rad(self):
        """
        Gives the outer radius (if circular) or radii (if elliptical - semi-major, semi-minor) of the
        region in which this set of light curves was generated.

        :return: The outer radius(ii) of the region.
        :rtype: Quantity
        """
        return self.all_lightcurves[0].outer_rad

    @property
    def time_bin_size(self) -> Quantity:
        """
        Gives the time bin size used to generate this set of light curves.

        :return: The time bin size used to generate this set of light curves.
        :rtype: Quantity
        """
        return self.all_lightcurves[0].time_bin_size

    @property
    def time_chunk_ids(self) -> np.ndarray:
        """
        Getter for the time chunk IDs associated with this AggregateLightCurve. Light curves that are part of an
        AggregateLightCurve will be separated into 'time chunks', where a time chunk is a period that has
        uninterrupted coverage. For instance, three XMM observations separated by a year each would be in three
        different time chunks, but if there were a fourth observation that was taken by another telescope and
        happened concurrently (even if it didn't start and end at the same time) with the first XMM observation, then
        it would be in the same time chunk.

        :return: np.ndarray
        :rtype: An array of integer time chunk identifiers, ordered from earlier to later times.
        """
        return self._time_chunk_ids

    @property
    def num_time_chunks(self) -> int:
        """
        Getter for the number of time chunks associated with this AggregateLightCurve. Light curves that are part
        of an AggregateLightCurve will be separated into 'time chunks', where a time chunk is a period that has
        uninterrupted coverage. For instance, three XMM observations separated by a year each would be in three
        different time chunks, but if there were a fourth observation that was taken by another telescope and
        happened concurrently (even if it didn't start and end at the same time) with the first XMM observation, then
        it would be in the same time chunk.

        :return: np.ndarray
        :rtype: An array of integer time chunk identifiers, ordered from earlier to later times.
        """
        return len(self._time_chunk_ids)

    @property
    def time_chunks(self) -> Quantity:
        """
        A getter for the start and stop times of the time chunks associated with this AggregateLightCurve. The left
        hand column are start times, and the right hand column are stop times. These are the earliest and latest
        times of coverage for all the observations in the particular time chunk.

        :return: A Nx2 non-scalar Astropy Quantity, where the left hand column are chunk start times, and the
            right hand column are chunk stop times.
        :rtype: Quantity
        """
        # Stores the chunk start and stop times until they are turned into a non-scalar Quantity at the end
        chunk_bounds = []
        # Iterates through the time chunks associated with this object
        for tc_id in self.time_chunk_ids:
            # Fetching lightcurves, making sure the return is a list
            rel_lcs = self.get_lightcurves(tc_id)
            if isinstance(rel_lcs, LightCurve):
                rel_lcs = [rel_lcs]
                # Finds the minimum start_time (i.e. the earliest start time) for lightcurves in this time chunk, and
                #  the maximum (i.e. latest) stop time - these define the bounds of this chunk
            tc_start = min(Quantity([lc.start_time for lc in rel_lcs]))
            tc_end = max(Quantity([lc.stop_time for lc in rel_lcs]))
            chunk_bounds.append(Quantity([tc_start, tc_end]))

        return Quantity(chunk_bounds)

    @property
    def datetime_chunks(self) -> np.ndarray:
        """
        A getter for the start and stop datetimes of the time chunks associated with this AggregateLightCurve. The
        left hand column are start datetimes, and the right hand column are stop datetimes. These are the earliest
        and latest times of coverage for all the observations in the particular time chunk.

        :return: A Nx2 array of datetime objects, where the left hand column are chunk start datetimes, and the
            right hand column are chunk stop datetimes.
        :rtype: np.ndarray(datetime)
        """
        # This behaves exactly the same as time_chunks, but works on datetime objects instead.
        chunk_bounds = []
        for tc_id in self.time_chunk_ids:
            rel_lcs = self.get_lightcurves(tc_id)
            if isinstance(rel_lcs, LightCurve):
                rel_lcs = [rel_lcs]
            tc_start = min([lc.start_datetime for lc in rel_lcs])
            tc_end = max([lc.stop_datetime for lc in rel_lcs])
            chunk_bounds.append([tc_start, tc_end])

        return np.array(chunk_bounds)

    @property
    def storage_key(self) -> str:
        """
        This property returns the storage key which this object assembles to place the AggregateLightCurve in
        an XGA source's storage structure. The key is based on the properties of the AggregateLightCurve, and
        some of the configuration options, and is basically human-readable.

        :return: String storage key.
        :rtype: str
        """
        return self._storage_key

    @property
    def event_selection_patterns(self) -> dict:
        """
        The event selection patterns used for different instruments that are associated with this AggregateLightCurve.

        :return: A dictionary where keys are instrument names and values are event selection patterns.
        :rtype: dict
        """
        return self._patterns

    # Then define user-facing methods
    def get_lightcurves(self, time_chunk_id: int, obs_id: str = None,
                        inst: str = None) -> Union[List[LightCurve], LightCurve]:
        """
        This is the getter for the lightcurves stored in the AggregateLightCurve data storage structure. They can
        be retrieved based on ObsID and instrument.

        :param int time_chunk_id: The time chunk identifier to retrieve lightcurves for.
        :param str obs_id: Optionally, a specific obs_id to search for can be supplied.
        :param str inst: Optionally, a specific instrument to search for can be supplied.
        :return: List of matching lightcurves, or just a LightCurve object if one match is found.
        :rtype: Union[List[LightCurve], LightCurve]
        """
        def unpack_list(to_unpack: list):
            """
            A recursive function to go through every layer of a nested list and flatten it all out. It
            doesn't return anything because to make life easier the 'results' are appended to a variable
            in the namespace above this one.

            :param list to_unpack: The list that needs unpacking.
            """
            # Must iterate through the given list
            for entry in to_unpack:
                # If the current element is not a list then all is chill, this element is ready for appending
                # to the final list
                if not isinstance(entry, list):
                    out.append(entry)
                else:
                    # If the current element IS a list, then obviously we still have more unpacking to do,
                    # so we call this function recursively.
                    unpack_list(entry)

        if time_chunk_id not in self.time_chunk_ids:
            tc_str = ", ".join(self.time_chunk_ids.astype(str))
            raise IndexError("{i} is not a time chunk ID associated with this AggregateLightCurve object. "
                             "Allowed time chunk IDs are; {a}".format(i=time_chunk_id, a=tc_str))
        elif obs_id not in self.obs_ids and obs_id is not None:
            raise NotAssociatedError("{0} is not associated with chunk {1} of this "
                                     "AggregateLightCurve.".format(obs_id, time_chunk_id))
        elif (obs_id is not None and obs_id in self.obs_ids) and \
                (inst is not None and inst not in self.instruments[obs_id]):
            raise NotAssociatedError("Instrument {1} is not associated with {0} for time chunk {2} of this "
                                     "AggregateLightCurve.".format(obs_id, inst, time_chunk_id))

        matches = []
        for match in dict_search(time_chunk_id, self._component_products):
            out = []
            unpack_list(match)
            if (obs_id == out[0] or obs_id is None) and (inst == out[1] or inst is None):
                matches.append(out[-1])

        # Here I only return the object if one match was found
        if len(matches) == 1:
            matches = matches[0]
        elif len(matches) == 0:
            # This probably means that an instrument was requested without a specific ObsID and it doesn't exist for
            #  the specified time chunk
            raise NotAssociatedError("The requested data are not associated with time chunk {} of this "
                                     "AggregateLightCurve.".format(time_chunk_id))
        return matches

    def get_data(self, inst: str, date_time: bool = False) -> Tuple[Quantity, Quantity, Union[TimeDelta, np.ndarray]]:
        """
        A get method to retrieve all count-rate and timing data for a particular instrument from this
        AggregateLightCurve. The data are in the correct temporal order.

        :param str inst: The instrument for which to retrieve the overall count-rate and time data.
        :param bool date_time: Whether the time data should be returned as an array of datetimes (not the default), or
            an Astropy TimeDelta object with the time as a different from MJD 50814.0 in seconds (the default).
        :return: The count rate data, count rate uncertainty data, and time data for the selected instrument. These
            are in the correct temporal order.
        :rtype: Tuple[Quantity, Quantity, Union[TimeDelta, np.ndarray]]
        """
        # These store the countrates, errors, and times that we pull out for the chosen instrument for all
        #  time chunks
        cr_data = []
        cr_err_data = []
        t_data = []
        # Iterate through the time chunk IDs associated with this object
        for tc_id in self.time_chunk_ids:
            try:
                # Grab the light curves, but catch if there isn't an entry for the chosen instrument for this
                #  time chunk and handle it gracefully
                rel_lcs = self.get_lightcurves(tc_id, inst=inst)
            except NotAssociatedError:
                continue

            # Append the current time chunk's chosen instrument's count rate data and error to their lists
            cr_data.append(rel_lcs.count_rate)
            cr_err_data.append(rel_lcs.count_rate_err)

            # Do the same with the datetime
            cur_dt = rel_lcs.datetime
            t_data.append(cur_dt)

        # Combine the datetime arrays in t_data into one array
        t_data = np.concatenate(t_data)
        # If the user wants the time data as a TimeDelta from the reference MJD time then calculate that
        if not date_time:
            t_data = (Time(t_data) - Time(50814.0, format='mjd')).sec

        # Concatenate the count rate data and error into one quantity each and return everything
        return np.concatenate(cr_data), np.concatenate(cr_err_data), t_data

    def get_view(self, fig: Figure, inst: str = None, custom_title: str = None, label_font_size: int = 18,
                 title_font_size: int = 20) -> Tuple[dict, Figure]:
        """
        A get method for a populated visualisation of the light curves present in this AggregateLightCurve.

        :param Figure fig: The matplotlib Figure object to create axes on, and thus make the plot.
        :param str inst: A specific instrument to display data for. Default is None, in which case all instruments
            are plotted.
        :param str custom_title: A custom title to add to the visualisation - the default is None, in which case a
            title containing the source name and energy band will be generated.
        :param int label_font_size: The font size for axes labels, default is 18.
        :param int title_font_size: The font size for the title, default is 20.
        :return: A dictionary of axes objects that have been added, and the figure object that was passed in.
        :rtype: Tuple[dict, Figure]
        """
        # TODO this will need a little bit of TLC once this and the multi-mission branch cross paths

        # This sets the fraction of the total x-width of the figure that is set between each axes
        buffer_frac = 0.008
        # This calculates the total time length of all time chunks
        chunk_len = (self.time_chunks[:, 1] - self.time_chunks[:, 0]).value
        # Then finds what fraction of the total coverage each time chunk covers, taking into account the buffer
        chunk_frac = chunk_len / (chunk_len.sum() + buffer_frac*len(chunk_len)-1)

        # Proportion of vertical to horizontal extent of the slanted line that breaks the separate axes
        break_slant = 1.3
        # Set properties of the break lines
        break_kwargs = dict(marker=[(-1, -break_slant), (1, break_slant)], markersize=12,
                            linestyle="none", color='k', mec='k', mew=1, clip_on=False)

        # To store the to-be-setup axes in
        axes_dict = {}
        # This is added to each iteration so that the next axes knows what x-position to start at
        cumu_x_pos = 0
        # Iterate through the time chunks, each will have a sub-axes
        for tc_id in self.time_chunk_ids:
            # Grab the fraction of time coverage for this time chunk (e.g. size of this axes)
            rel_frac = chunk_frac[tc_id]

            # We deal with axes differently depending on whether they are the first time chunk in a series, one in
            #  the middle of a series of time chunks, or the last. We also have to account for the fact that there
            #  could only be one or two time chunks. This all defines whether we plot break lines, which axes
            #  boundaries are visible, which axis tick labels are shown etc.
            if tc_id == 0:
                # Adding a new axes, the size defined by the fractional coverage time, and the position by
                #  cumu_x_pos (though that should always be zero for tc_id == 0). It extends the full height of
                #  the figure
                axes_dict[tc_id] = fig.add_axes([cumu_x_pos, 0.0, rel_frac, 1])
                # Set up a y-label - other axes won't have this because they share the y-axis and we only want to
                #  label the first one
                y_lab = "Count-rate [{}]".format(self.all_lightcurves[0].count_rate.unit.to_string('latex'))
                axes_dict[tc_id].set_ylabel(y_lab, fontsize=label_font_size)

                # We set the upper and lower y-axis limits based on the maximum and minimum count rates across all
                #  the lightcurves in this object, as the y-axis is shared
                low_lim = min([(lc.count_rate-lc.count_rate_err).min() for lc in self.all_lightcurves]).value*0.95
                upp_lim = max([(lc.count_rate+lc.count_rate_err).max() for lc in self.all_lightcurves]).value*1.05
                axes_dict[tc_id].set_ylim(low_lim, upp_lim)

                # If there is more than one time chunk, we turn off the line on the right hand side of this initial
                #  axes - so there is no unsightly barrier between it and the next axes - we also add slanted lines
                #  to indicate a break in the y-axis
                if self.num_time_chunks != 1:
                    axes_dict[tc_id].spines.right.set_visible(False)
                    axes_dict[tc_id].plot([1, 1], [1, 0], transform=axes_dict[tc_id].transAxes, **break_kwargs)

                # We make sure the ticks look how we want them
                axes_dict[tc_id].tick_params(direction='in', which='both', right=False, left=True, top=True)

            # In this case we are at a time chunk that is not the first, and not the last
            elif tc_id != (self.num_time_chunks - 1):
                # Add the axes at the correct position, making sure to share the y-axis with the first axes
                axes_dict[tc_id] = fig.add_axes([cumu_x_pos, 0.0, rel_frac, 1], sharey=axes_dict[0])
                # Both the left hand and the right hand axis lines are turned off
                axes_dict[tc_id].spines.left.set_visible(False)
                axes_dict[tc_id].spines.right.set_visible(False)
                # We make sure to setup the ticks as we want them - making sure that the y-axis is not labelled for
                #  this one, and that the left and right ticks are turned off
                axes_dict[tc_id].tick_params(direction='in', which='both', right=False, left=False, top=True,
                                             labelleft=False)
                # This is what sets up the break lines on the left and right hand sides of the x-axis, at the top
                #  and bottom
                axes_dict[tc_id].plot([1, 1], [1, 0], transform=axes_dict[tc_id].transAxes, **break_kwargs)
                axes_dict[tc_id].plot([0, 0], [0, 1], transform=axes_dict[tc_id].transAxes, **break_kwargs)

            # Finally this is triggered when we're at the last time chunks
            else:
                axes_dict[tc_id] = fig.add_axes([cumu_x_pos, 0.0, rel_frac, 1], sharey=axes_dict[0])
                # The right hand axis line is visible, but not the left
                axes_dict[tc_id].spines.right.set_visible(True)
                axes_dict[tc_id].spines.left.set_visible(False)
                # And we make sure to add the tick setup and ensure that the final break lines on the left are drawn
                axes_dict[tc_id].tick_params(direction='in', which='both', right=True, left=False, top=True,
                                             labelleft=False)
                axes_dict[tc_id].plot([0, 0], [0, 1], transform=axes_dict[tc_id].transAxes, **break_kwargs)

            # Iterate the cumulative position
            cumu_x_pos += (rel_frac+buffer_frac)
            # And turn on minor ticks, because I prefer how that looks
            axes_dict[tc_id].minorticks_on()

            # Setting the x-axis limits, based on the known time coverage of the time chunk
            axes_dict[tc_id].set_xlim(self.datetime_chunks[tc_id, 0], self.datetime_chunks[tc_id, 1])

        # Create a single x-axis label for all axes, it looks ugly and is unnecessary to have one for each
        fig.text(0.5, -0.04, "Time", ha='center', fontsize=label_font_size)

        # Now we need to populate our carefully set up axes with DATA
        for tc_id in self.time_chunk_ids:
            ax = axes_dict[tc_id]
            try:
                # That this is the first part of the try-except, and won't trigger the except, is quite deliberate. It
                #  sets up the x-axis tick labels, even if there are no data for the requested instrument (if the
                #  user has even requested a specific instrument) - makes the plot look nice
                # The labels will have hour, minute, date, month (shortened word so as not to be ambigious with
                #  american date system), and year
                ax.xaxis.set_major_formatter(mdates.DateFormatter('%Hh-%Mm %d-%b-%Y'))
                for label in ax.get_xticklabels(which='major'):
                    label.set(y=label.get_position()[1] - 0.03, rotation=40, horizontalalignment='right')

                # This is what _might_ trigger the NotAssociatedError
                rel_lcs = self.get_lightcurves(tc_id, inst=inst)
            except NotAssociatedError:
                continue

            # Make sure rel_lcs, even if it only contains a single light curve
            if isinstance(rel_lcs, LightCurve):
                rel_lcs = [rel_lcs]

            # Now we cycle through the light curves for the current time chunk and add them to the plot
            for rel_lc in rel_lcs:
                ident = "{t} {o}-{i}".format(t='XMM', o=rel_lc.obs_id, i=rel_lc.instrument)
                ax.errorbar(rel_lc.datetime, rel_lc.count_rate.value, yerr=rel_lc.count_rate_err.value,
                            capsize=2, label=ident, fmt='x')

            ax.legend(loc='best')

        # Check if the user has defined a custom title, and if not then we build one and add it to the plot
        if custom_title is not None:
            fig.suptitle(custom_title, fontsize=title_font_size, y=1.05)
        elif self.src_name is not None:
            fig.suptitle("{s} {l}-{u}keV Aggregate Lightcurve".format(s=self.src_name,
                                                                      l=self.energy_bounds[0].to('keV').value,
                                                                      u=self.energy_bounds[1].to('keV').value),
                         fontsize=title_font_size, y=1.05)
        else:
            fig.suptitle("{l}-{u}keV Aggregate Lightcurve".format(l=self.energy_bounds[0].to('keV').value,
                                                                  u=self.energy_bounds[1].to('keV').value),
                         fontsize=title_font_size, y=1.05)

        return axes_dict, fig

    def view(self, figsize: tuple = (14, 6), inst: str = None, custom_title: str = None, label_font_size: int = 15,
             title_font_size: int = 18):
        """
        This method creates a combined visualisation of all the light curves associated with this object (apart from
        when you specify a single instrument, then it uses all the light curves from that instrument). The data are
        displayed in the correct temporal order, with the x-axis labels indicating the date and time rather than the
        mission specific internal time.

        :param tuple figsize: The size of the visualisation figure, default is (14, 6). Adjusting this value is the
            best way to achieve nice looking plots when labels are overlapping, particularly when there are many
            observations and time chunks to plot in the x-direction.
        :param str inst: A specific instrument to display data for. Default is None, in which case all instruments
            are plotted.
        :param str custom_title: A custom title to add to the visualisation - the default is None, in which case a
            title containing the source name and energy band will be generated.
        :param int label_font_size: The font size for axes labels, default is 18.
        :param int title_font_size: The font size for the title, default is 20.
        """
        # Create figure object
        fig = plt.figure(figsize=figsize)

        ax_dict, fig = self.get_view(fig, inst, custom_title, label_font_size, title_font_size)

        # plt.tight_layout()
        # Display the plot
        plt.show()

        # Wipe the figure
        plt.close("all")



#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 20/02/2023, 14:04. Copyright (c) The Contributors

import os
import warnings
from multiprocessing.dummy import Pool
from typing import Tuple, Union

import numpy as np
from astropy.units import Quantity
from fitsio import FITSHDR
from fitsio import write
from scipy.signal import convolve
from tqdm import tqdm

from ..products import PSFGrid, Image
from ..samples.base import BaseSample
from ..sas import evselect_image, psfgen, emosaic
from ..sources import BaseSource
from ..utils import NUM_CORES, OUTPUT


def rl_psf(sources: Union[BaseSource, BaseSample], iterations: int = 15, psf_model: str = "ELLBETA",
           lo_en: Quantity = Quantity(0.5, 'keV'), hi_en: Quantity = Quantity(2.0, 'keV'),
           bins: int = 4, num_cores: int = NUM_CORES):
    """
    An implementation of the Richardson-Lucy (doi:10.1364/JOSA.62.000055) PSF deconvolution algorithm that
    also takes into account the spatial variance of the XMM Newton PSF. The sources passed into this
    function will have all images matching the passed energy range deconvolved, the image objects will have the
    result stored in them alongside the original data, and a combined image will be generated. I view this
    method as quite crude, but it does seem to work, and I may implement a more intelligent way of doing
    PSF deconvolutions later.
    I initially tried convolving the PSFs generated at different spatial points with the chunks of data relevant
    to them in isolation, but the edge effects were very obvious. So I settled on convolving the whole
    original image with each PSF, and after it was finished taking the relevant chunks and patchworking
    them into a new array.

    :param BaseSource/BaseSample sources: A single source object, or list of source objects.
    :param int iterations: The number of deconvolution iterations performed by the Richardson-Lucy algorithm.
    :param str psf_model: Which model of PSF should be used for this deconvolution. The default is ELLBETA,
        the best available.
    :param Quantity lo_en: The lower energy bound of the images to be deconvolved.
    :param Quantity hi_en: The upper energy bound of the images to be deconvolved.
    :param int bins: Number of bins that the X and Y axes will be divided into when generating a PSFGrid.
    :param int num_cores: The number of cores to use (if running locally), the default is set to 90%
        of available cores in your system.
    """
    def rl_step(ind: int, cur_image: np.ndarray, last_image: np.ndarray, rel_psf: np.ndarray) \
            -> Tuple[np.ndarray, int]:
        """
        This performs one iteration of the Richardson-Lucy PSF deconvolution method. Basically copied from
        the skimage implementation, but set up so that we can multiprocess this, as well as do it in steps
        and save each step separately.

        :param int ind: The current step, passed through for the callback function.
        :param np.ndarray cur_image: The im_deconv from the last step.
        :param last_image:
        :param rel_psf: The particular spacial PSF being applied to this image.
        :return:
        """
        psf_mirror = rel_psf[::-1, ::-1]
        relative_blur = cur_image / convolve(last_image, rel_psf, mode='same')
        im_deconv = last_image * convolve(relative_blur, psf_mirror, mode='same')

        return im_deconv, ind

    def new_header(og_header: FITSHDR) -> FITSHDR:
        """
        Modifies an existing XMM Newton fits image header, removes some elements, and adds a little extra
        information. The new header is then used for PSF corrected fits image files.

        :param og_header: The header from the fits image that has been PSF corrected.
        :return: The new, modified, fits header.
        :rtype: FITSHDR
        """
        # These are nuisance entries that I want gone
        remove_list = ["CREATOR", "CONTINUE", "XPROC0", "XDAL0"]

        # FITSHDR takes a dictionary as its main content argument
        new_header_info = {}
        # We want to copy most of the original header
        for e in og_header:
            # But those that are in the remove list aren't copied over.
            if e not in remove_list:
                new_header_info[e] = og_header[e]
            # Also change the creator entry for a little self serving brand placement
            elif e == "CREATOR":
                new_header_info[e] = "XGA"

        # Setting some new headers for further information
        new_header_info["COMMENT"] = "THIS IMAGE HAS BEEN PSF CORRECTED BY XGA"
        new_header_info["PSFBins"] = bins
        new_header_info["PSFAlgorithm"] = "Richardson-Lucy"
        new_header_info["PSFAlgorithmIterations"] = iterations
        new_header_info["PSFModel"] = psf_model

        return FITSHDR(new_header_info)

    # These just make sure that the images to be deconvolved and the PSFs to deconvolve them
    #  with have actually been generated. If they are already associated with the sources then this won't
    #  do anything.
    lo_en = lo_en.to('keV')
    hi_en = hi_en.to('keV')
    # Making sure that the necessary images and PSFs are generated
    evselect_image(sources, lo_en, hi_en, num_cores=num_cores)

    # If just one source is passed in, make it a list of one, makes behaviour more consistent
    #  throughout this function.
    if not isinstance(sources, (list, BaseSample)):
        sources = [sources]

    # Only those sources that don't already have the individual PSF corrected images should be run
    sub_sources = []
    for source in sources:
        en_id = "bound_{l}-{u}".format(l=lo_en.value, u=hi_en.value)
        # All the image objects of the specified energy range (so every combination of ObsID and instrument)
        match_images = source.get_products("image", extra_key=en_id)

        # This is the key under which the PSF corrected image will be stored, defining it to check that
        #  it doesn't already exist.
        key = "bound_{l}-{u}_{m}_{b}_rl{i}".format(l=float(lo_en.value), u=float(hi_en.value), m=psf_model,
                                                   b=bins, i=iterations)
        # Check to see if all individual PSF corrected images are present
        psf_corr_prod = [p for p in source.get_products("image", just_obj=False) if key in p]

        # If all the PSF corrected images are present then we skip, the correction has already been performed.
        if len(psf_corr_prod) == len(match_images):
            continue
        else:
            sub_sources.append(source)

    # Should have cleaned it so that only those sources that need it will have PSFs generated
    psfgen(sub_sources, bins, psf_model, num_cores=num_cores)

    corr_prog_message = 'PSF Correcting Observations - Currently {}'
    with tqdm(desc=corr_prog_message.format(''), total=len(sub_sources),
              disable=len(sub_sources) == 0) as corr_progress:
        for source in sub_sources:
            source: BaseSource
            # Updates the source name in the message every iteration
            corr_progress.set_description(corr_prog_message.format(source.name))

            en_id = "bound_{l}-{u}".format(l=lo_en.value, u=hi_en.value)
            # All the image objects of the specified energy range (so every combination of ObsID and instrument)
            match_images = source.get_products("image", extra_key=en_id)

            # Just warns the user that some of the images may not be valid
            for matched in match_images:
                if "PrimeFullWindow" not in matched.header["SUBMODE"]:
                    warnings.warn("PSF corrected images for {s}-{o}-{i} may not be valid, as the data was taken"
                                  "in {m} mode".format(s=source.name, o=matched.obs_id, i=matched.instrument,
                                                       m=matched.header["SUBMODE"]))

            # For now just going to iterate through them, we'll see if I can improve it later
            for im in match_images:
                # Read these out just because they'll be useful
                obs_id = im.obs_id
                inst = im.instrument

                # Extra key we need to search for the PSFGrid we need, then fetch it.
                psf_key = "_".join([psf_model, str(bins)])
                psf_grid: PSFGrid = source.get_products("psf", obs_id, inst, psf_key)[0]

                # This uses a built in method of the PSF class to re-sample the PSF(s) to the same
                #  scale as our image
                resamp_psfs = []
                for psf in psf_grid:
                    resamp_psfs.append(psf.resample(im, Quantity(64, 'pix')))

                full_im_data = im.data.copy()
                # This adds very low level random noise to the image, getting rid of 0 values, this makes life
                #  easier with the convolution functions
                full_im_data += full_im_data.max() * 1E-8 * np.random.random(full_im_data.shape)

                # Nasty nested list to store each step for each PSF in
                # The starting array full of 0.5 values is what is used in the skimage implementation
                storage = [[np.full(full_im_data.shape, 0.5)] for i in range(len(psf_grid))]
                # Iterating over the steps
                for i in range(iterations):
                    # Sets up a multiprocessing pool
                    with Pool(num_cores) as pool:
                        # This is what is triggered when each rl_step call is done
                        def callback(results_in):
                            nonlocal storage
                            # Reading out results
                            proc_chunk, stor_ind = results_in
                            # Storing the results in the storage listss
                            storage[stor_ind].append(proc_chunk)

                        def err_callback(err):
                            raise err

                        # Starts a deconvolution function for each spatial PSF, adding them to the
                        #  multiprocessing pool
                        for psf_ind, psf in enumerate(psf_grid):
                            pool.apply_async(rl_step, callback=callback, error_callback=err_callback,
                                             args=(psf_ind, full_im_data, storage[psf_ind][-1], resamp_psfs[psf_ind]))
                        pool.close()  # No more tasks can be added to the pool
                        pool.join()  # Joins the pool, the code will only move on once the pool is empty.

                # Because of the nature of convolutions, you can end up with a significantly
                #  different sum total for the image than you started with. To maintain physicality, we need to
                #  preserve the total number of counts, so we re-normalise each step using the original total.
                og_total = im.data.sum()
                # The 3D array that the final patch-worked images are stored in
                # Every iteration has its resultant image stored in here, earliest at index 0, final at index -1
                final_form = np.zeros((im.shape[0], im.shape[1], iterations))
                # Iterate over the PSFs generated at different points
                for psf_ind, psf in enumerate(psf_grid):
                    del storage[psf_ind][0]  # You may remember this was just the initial array of 0.5 as per skimage
                    # Change from an N, 512, 512 (for XGA images) to 512, 512, N array (makes more sense to my brain).
                    deconv_steps = np.moveaxis(np.array(storage[psf_ind]), 0, 2)

                    # Grab out the x and y boundary coordinates for the spatial chunk relevant to the current PSF
                    x_lims = psf_grid.x_bounds[psf_ind, :]
                    y_lims = psf_grid.y_bounds[psf_ind, :]

                    # For each deconvolution step we find the total value of the image and normalise by
                    #  the original total
                    step_totals = np.sum(deconv_steps, (0, 1))
                    norm_factors = og_total / step_totals
                    deconv_steps = deconv_steps * norm_factors

                    # Cut out the piece of the image array (for all iteration steps) that is relevant for the current
                    # PSF and insert it into the final image array at the same coordinates
                    # Patch-working yay, terminology definitely inspired by the new Passenger Album :)
                    final_form[y_lims[0]:y_lims[1], x_lims[0]:x_lims[1], :] = \
                        deconv_steps[y_lims[0]:y_lims[1], x_lims[0]:x_lims[1], :]

                # Final re-normalization, bit cheesy but oh well
                norm_factors = og_total / np.sum(final_form, (0, 1))
                final_form = final_form * norm_factors

                # Define file names for the whole datacube (all iteration steps), and the final image (just the last).
                datacube_name = "{o}_{i}_{b}bin_{it}iter_{m}mod_rlalgo_{l}-{u}keVpsfcorr_datacube." \
                                "fits".format(o=obs_id, i=inst, b=bins, it=iterations, l=lo_en.value, u=hi_en.value,
                                              m=psf_model)
                # Define this one because emosaic can't deal with fits datacubes, and thats what I use to combine images.
                im_name = "{o}_{i}_{b}bin_{it}iter_{m}mod_rlalgo_{l}-{u}keVpsfcorr_img." \
                          "fits".format(o=obs_id, i=inst, b=bins, it=iterations, l=lo_en.value, u=hi_en.value,
                                        m=psf_model)

                # Use the super handy fitsio write function to create the new fits datacube, and final image.
                write(os.path.join(OUTPUT, obs_id, datacube_name), np.moveaxis(final_form, 2, 0),
                      header=new_header(im.header))
                write(os.path.join(OUTPUT, obs_id, im_name), np.moveaxis(final_form, 2, 0)[-1, :, :],
                      header=new_header(im.header))

                # Makes an XGA product of our brand new image
                fin_im = Image(os.path.join(OUTPUT, obs_id, im_name), obs_id, inst, '', '', '', lo_en, hi_en)
                # Adds PSF correction information for XGA's internal use
                fin_im.psf_corrected = True
                fin_im.psf_algorithm = "rl"
                fin_im.psf_bins = bins
                fin_im.psf_iterations = iterations
                fin_im.psf_model = psf_model

                # Adds it into the source
                source.update_products(fin_im)

                # Removes the PSFGrid constituents from memory
                psf_grid.unload_data()

            corr_progress.update(1)
        corr_progress.set_description(corr_prog_message.format('complete'))

    # For the passed sources we now run emosaic to create combined PSF corrected images.
    emosaic(sources, "image", lo_en, hi_en, psf_corr=True, psf_model=psf_model, psf_bins=bins,
            psf_algo="rl", psf_iter=iterations)

